from __future__ import annotations

import builtins
from collections.abc import Generator, Iterator
from dataclasses import dataclass
from io import StringIO
import os
from types import SimpleNamespace
from typing import Any, Literal, cast

from prompt_toolkit.formatted_text import to_formatted_text, to_plain_text
import pymysql
import pytest

import mycli.main_modes.repl as repl_mode
from mycli.packages.sqlresult import SQLResult


class DummyLogger:
    def __init__(self) -> None:
        self.debug_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []
        self.error_calls: list[tuple[tuple[Any, ...], dict[str, Any]]] = []

    def debug(self, *args: Any, **kwargs: Any) -> None:
        self.debug_calls.append((args, kwargs))

    def error(self, *args: Any, **kwargs: Any) -> None:
        self.error_calls.append((args, kwargs))


class HashableNamespace:
    pass


@dataclass
class DummyFormatterWithQuery:
    query: str = ''


class FakeApp:
    def __init__(self, text: str = '', render_counter: int = 0) -> None:
        self.current_buffer = SimpleNamespace(text=text)
        self.render_counter = render_counter
        self.ttimeoutlen: float | None = None


class FakePromptOutput:
    def __init__(self, columns: int = 80, rows: int = 24) -> None:
        self.columns = columns
        self.rows = rows
        self.bell_count = 0

    def get_size(self) -> SimpleNamespace:
        return SimpleNamespace(columns=self.columns, rows=self.rows)

    def bell(self) -> None:
        self.bell_count += 1


class FakePromptSession:
    def __init__(self, responses: list[Any] | None = None, columns: int = 80, rows: int = 24) -> None:
        self.responses = list(responses or [])
        self.output = FakePromptOutput(columns=columns, rows=rows)
        self.app = FakeApp()
        self.prompt_calls: list[dict[str, Any]] = []

    def prompt(self, **kwargs: Any) -> str:
        self.prompt_calls.append(dict(kwargs))
        if not self.responses:
            raise EOFError()
        response = self.responses.pop(0)
        if isinstance(response, BaseException):
            raise response
        return response


class FakeCursorBase:
    def __init__(
        self,
        rows: list[tuple[Any, ...]] | None = None,
        rowcount: int = 0,
        warning_count: int = 0,
    ) -> None:
        self._rows = list(rows or [])
        self.rowcount = rowcount
        self.warning_count = warning_count

    def __iter__(self) -> Iterator[tuple[Any, ...]]:
        return iter(self._rows)


class FakeConnection:
    def __init__(self, ping_exc: Exception | None = None, cursor_value: Any = 'cursor') -> None:
        self.ping_exc = ping_exc
        self.cursor_value = cursor_value
        self.ping_calls: list[bool] = []

    def ping(self, reconnect: bool = False) -> None:
        self.ping_calls.append(reconnect)
        if self.ping_exc is not None:
            raise self.ping_exc

    def cursor(self) -> Any:
        return self.cursor_value


class ReusableLock:
    def __enter__(self) -> 'ReusableLock':
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> Literal[False]:
        return False


def sqlresult_generator(*results: SQLResult) -> Generator[SQLResult, None, None]:
    for result in results:
        yield result


class FakeResourceTree:
    def __init__(self, files: dict[str, str], path: str | None = None) -> None:
        self.files = files
        self.path = path

    def joinpath(self, path: str) -> 'FakeResourceTree':
        return FakeResourceTree(self.files, path)

    def open(self, mode: str) -> StringIO:
        assert self.path is not None
        if self.path not in self.files:
            raise FileNotFoundError(self.path)
        return StringIO(self.files[self.path])


def make_repl_cli(sqlexecute: Any | None = None) -> Any:
    cli: Any = HashableNamespace()
    cli.logger = DummyLogger()
    cli.query_history = []
    cli.last_prompt_message = repl_mode.ANSI('')
    cli.last_custom_toolbar_message = repl_mode.ANSI('')
    cli.prompt_lines = 0
    cli.default_prompt = r'\t \u@\h:\d> '
    cli.default_prompt_splitln = r'\u@\h\n(\t):\d>'
    cli.max_len_prompt = 45
    cli.prompt_format = cli.default_prompt
    cli.multiline_continuation_char = '>'
    cli.toolbar_format = 'default'
    cli.verbosity = -1
    cli.keepalive_ticks = None
    cli._keepalive_counter = 0
    cli.auto_vertical_output = False
    cli.beep_after_seconds = 0.0
    cli.show_warnings = False
    cli.null_string = '<null>'
    cli.numeric_alignment = 'right'
    cli.binary_display = None
    cli.prompt_session = None
    cli.post_redirect_command = None
    cli.logfile = None
    cli.smart_completion = False
    cli.config = {'main': {'history_file': '~/.mycli-history-testing'}}
    cli.key_bindings = 'emacs'
    cli.wider_completion_menu = False
    cli.login_path = None
    cli.login_path_as_host = False
    cli.dsn_alias = None
    cli.terminal_tab_title_format = ''
    cli.terminal_window_title_format = ''
    cli.multiplex_window_title_format = ''
    cli.multiplex_pane_title_format = ''
    cli._completer_lock = ReusableLock()
    cli.completer = object()
    cli.syntax_style = 'native'
    cli.cli_style = {}
    cli.emacs_ttimeoutlen = 1.0
    cli.vi_ttimeoutlen = 2.0
    cli.sandbox_mode = False
    cli.destructive_warning = False
    cli.destructive_keywords = ['drop']
    cli.llm_prompt_field_truncate = 0
    cli.llm_prompt_section_truncate = 0
    cli.main_formatter = DummyFormatterWithQuery()
    cli.redirect_formatter = DummyFormatterWithQuery()
    cli.pager_configured = 0
    refresh_calls: list[bool] = []
    output_calls: list[tuple[list[str], Any, bool]] = []
    echo_calls: list[str] = []
    timing_calls: list[tuple[str, bool]] = []
    log_queries: list[str] = []
    cli.refresh_calls = refresh_calls
    cli.output_calls = output_calls
    cli.echo_calls = echo_calls
    cli.timing_calls = timing_calls
    cli.log_queries = log_queries
    cli.title_calls = 0
    cli.sqlexecute = sqlexecute
    cli.get_reserved_space = lambda: 3
    cli.get_last_query = lambda: cli.query_history[-1].query if cli.query_history else None
    cli.configure_pager = lambda: setattr(cli, 'pager_configured', cli.pager_configured + 1)

    def refresh_completions(reset: bool = False) -> list[SQLResult]:
        cli.refresh_calls.append(reset)
        return [SQLResult(status='refresh')]

    cli.refresh_completions = refresh_completions

    def output_timing(timing: str, is_warnings_style: bool = False) -> None:
        cli.timing_calls.append((timing, is_warnings_style))

    cli.output_timing = output_timing

    def log_query(text: str) -> None:
        cli.log_queries.append(text)

    cli.log_query = log_query
    cli.reconnect = lambda database='': False

    def echo(message: Any, **kwargs: Any) -> None:
        cli.echo_calls.append(str(message))

    cli.echo = echo

    def format_sqlresult(result: SQLResult, **kwargs: Any) -> Iterator[str]:
        return iter([str(kwargs.get('max_width')), result.status_plain or 'row'])

    cli.format_sqlresult = format_sqlresult

    def output(formatted: Any, result: Any, is_warnings_style: bool = False) -> None:
        cli.output_calls.append((list(formatted), result, is_warnings_style))

    cli.output = output
    return cli


def patch_repl_runtime_defaults(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(repl_mode.special, 'set_expanded_output', lambda value: None)
    monkeypatch.setattr(repl_mode.special, 'set_forced_horizontal_output', lambda value: None)
    monkeypatch.setattr(repl_mode.special, 'is_llm_command', lambda text: False)
    monkeypatch.setattr(repl_mode.special, 'is_expanded_output', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'is_timing_enabled', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'write_tee', lambda *args, **kwargs: None)
    monkeypatch.setattr(repl_mode.special, 'unset_once_if_written', lambda *args, **kwargs: None)
    monkeypatch.setattr(repl_mode.special, 'flush_pipe_once_if_written', lambda *args, **kwargs: None)
    monkeypatch.setattr(repl_mode.special, 'close_tee', lambda: None)
    monkeypatch.setattr(repl_mode, 'handle_editor_command', lambda mycli, text, inputhook, loaded_message_fn: text)
    monkeypatch.setattr(repl_mode, 'handle_clip_command', lambda mycli, text: False)
    monkeypatch.setattr(repl_mode, 'is_redirect_command', lambda text: False)
    monkeypatch.setattr(repl_mode, 'confirm_destructive_query', lambda keywords, text: None)
    monkeypatch.setattr(repl_mode, 'need_completion_refresh', lambda text: False)
    monkeypatch.setattr(repl_mode, 'need_completion_reset', lambda text: False)
    monkeypatch.setattr(repl_mode, 'is_dropping_database', lambda text, dbname: False)
    monkeypatch.setattr(repl_mode, 'is_mutating', lambda status: False)


def test_complete_while_typing_filter_covers_threshold_and_word_rules(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(repl_mode, 'MIN_COMPLETION_TRIGGER', 3)
    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='ab')))
    assert repl_mode.complete_while_typing_filter() is False

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='abc')))
    assert repl_mode.complete_while_typing_filter() is True

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='source xyz')))
    assert repl_mode.complete_while_typing_filter() is True

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='source x/')))
    assert repl_mode.complete_while_typing_filter() is False

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='\\. abc')))
    assert repl_mode.complete_while_typing_filter() is True

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='\\. a/')))
    assert repl_mode.complete_while_typing_filter() is False

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='select abc')))
    assert repl_mode.complete_while_typing_filter() is True

    monkeypatch.setattr(repl_mode, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='select a!')))
    assert repl_mode.complete_while_typing_filter() is False

    monkeypatch.setattr(repl_mode, 'MIN_COMPLETION_TRIGGER', 1)
    assert repl_mode.complete_while_typing_filter() is True


def test_repl_create_history(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_repl_cli()
    monkeypatch.setenv('MYCLI_HISTFILE', '~/override-history')
    monkeypatch.setattr(repl_mode, 'dir_path_exists', lambda path: True)
    monkeypatch.setattr(repl_mode, 'FileHistoryWithTimestamp', lambda path: f'history:{path}')
    history = cast(Any, repl_mode._create_history(cli))
    assert history == f'history:{os.path.expanduser("~/override-history")}'

    monkeypatch.delenv('MYCLI_HISTFILE')
    monkeypatch.setattr(repl_mode, 'dir_path_exists', lambda path: False)
    assert repl_mode._create_history(cli) is None
    assert 'Unable to open the history file' in cli.echo_calls[-1]


def test_repl_picker_helpers_cover_present_and_missing_resources(monkeypatch: pytest.MonkeyPatch) -> None:
    files = {
        'AUTHORS': '* Alice\n* Bob\n',
        'SPONSORS': '* Carol\n',
        'TIPS': '# comment\nTip 1\n\nTip 2\n',
    }
    monkeypatch.setattr(repl_mode.resources, 'files', lambda package: FakeResourceTree(files))
    monkeypatch.setattr(repl_mode.random, 'choice', lambda seq: seq[0])
    assert repl_mode._contributors_picker() == 'Alice'
    assert repl_mode._sponsors_picker() == 'Carol'
    assert repl_mode._tips_picker() == 'Tip 1'

    monkeypatch.setattr(repl_mode.resources, 'files', lambda package: FakeResourceTree({}))
    assert repl_mode._contributors_picker() == 'our contributors'
    assert repl_mode._sponsors_picker() == 'our sponsors'
    assert repl_mode._tips_picker() == r'\? or "help" for help!'


def test_repl_show_startup_banner_and_prompt_helpers(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_repl_cli(SimpleNamespace(server_info='Server'))
    printed: list[str] = []
    monkeypatch.setattr(builtins, 'print', lambda *args, **kwargs: printed.append(' '.join(str(x) for x in args)))
    monkeypatch.setattr(repl_mode.random, 'random', lambda: 0.4)
    monkeypatch.setattr(repl_mode, '_contributors_picker', lambda: 'Alice')
    monkeypatch.setattr(repl_mode, '_sponsors_picker', lambda: 'Carol')
    monkeypatch.setattr(repl_mode, '_tips_picker', lambda: 'Tip')

    cli.verbosity = 0
    repl_mode._show_startup_banner(cli, cli.sqlexecute)
    monkeypatch.setattr(repl_mode.random, 'random', lambda: 0.6)
    repl_mode._show_startup_banner(cli, cli.sqlexecute)
    cli.verbosity = -1
    repl_mode._show_startup_banner(cli, cli.sqlexecute)
    assert any('Thanks to the contributor' in line for line in printed)
    assert any('Tip — Tip' in line for line in printed)

    monkeypatch.setattr(
        repl_mode,
        'render_prompt_string',
        lambda mycli, string, render_counter: to_formatted_text('0123456') if string == cli.default_prompt else 'a\nb',
    )
    cli.max_len_prompt = 5
    prompt_text = to_plain_text(repl_mode._get_prompt_message(cli, cast(Any, FakeApp(text='', render_counter=2))))
    assert prompt_text == 'a\nb'
    assert cli.prompt_lines == 2

    cli.last_prompt_message = repl_mode.ANSI('cached')
    assert to_plain_text(repl_mode._get_prompt_message(cli, cast(Any, FakeApp(text='typing', render_counter=3)))) == 'cached'

    cli.prompt_format = 'custom'
    cli.prompt_lines = 0
    monkeypatch.setattr(repl_mode, 'render_prompt_string', lambda mycli, string, render_counter: to_formatted_text('single'))
    assert to_plain_text(repl_mode._get_prompt_message(cli, cast(Any, FakeApp(text='', render_counter=4)))) == 'single'
    assert cli.prompt_lines == 1

    assert repl_mode._get_continuation(cli, 4, 0, 0) == [('class:continuation', '  > ')]
    cli.multiline_continuation_char = ''
    assert repl_mode._get_continuation(cli, 4, 0, 0) == [('class:continuation', '')]
    cli.multiline_continuation_char = None
    assert repl_mode._get_continuation(cli, 4, 0, 0) == [('class:continuation', ' ')]


def test_repl_show_startup_banner_thanks_sponsor(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_repl_cli(SimpleNamespace(server_info='Server'))
    cli.verbosity = 0
    printed: list[str] = []
    monkeypatch.setattr(builtins, 'print', lambda *args, **kwargs: printed.append(' '.join(str(x) for x in args)))
    monkeypatch.setattr(repl_mode.random, 'random', lambda: 0.25)
    monkeypatch.setattr(repl_mode, '_sponsors_picker', lambda: 'Carol')

    repl_mode._show_startup_banner(cli, cli.sqlexecute)

    assert any('Thanks to the sponsor' in line and 'Carol' in line for line in printed)


def test_prompt_toolbar_and_title_helpers(monkeypatch: pytest.MonkeyPatch) -> None:
    class PromptCursor:
        def __enter__(self) -> 'PromptCursor':
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> Literal[False]:
            return False

    class PromptConnection:
        def cursor(self) -> PromptCursor:
            return PromptCursor()

    sqlexecute = SimpleNamespace(
        user='alice',
        host='127.0.0.1',
        dbname='db',
        port=3307,
        socket='/tmp/mysql.sock',
        server_info=SimpleNamespace(species=SimpleNamespace(name='TiDB')),
        conn=None,
    )
    cli = make_repl_cli(sqlexecute)
    cli.login_path = 'prod'
    cli.login_path_as_host = True
    cli.dsn_alias = 'dsn'
    prompt = repl_mode.render_prompt_string(cli, r'\h|\H|\A|\y|\Y|\T|\w|\W', 0)
    prompt_plain = to_plain_text(prompt)
    assert prompt_plain == 'prod|prod|dsn|(none)|(none)|(none)|(none)|'

    sqlexecute.conn = PromptConnection()
    cli.login_path_as_host = False
    monkeypatch.setattr(repl_mode, 'get_uptime', lambda cur: 123)
    monkeypatch.setattr(repl_mode, 'format_uptime', lambda uptime: f'uptime:{uptime}')
    monkeypatch.setattr(repl_mode, 'get_ssl_version', lambda cur: 'TLSv1.3')
    monkeypatch.setattr(repl_mode, 'get_warning_count', lambda cur: 7)
    prompt = repl_mode.render_prompt_string(cli, r'\H|\y|\Y|\T|\w|\W', 1)
    prompt_plain = to_plain_text(prompt)
    assert prompt_plain == '127.0.0.1|123|uptime:123|TLSv1.3|7|7'

    cli.prompt_session = None
    assert to_plain_text(repl_mode.get_custom_toolbar(cli, 'fmt')) == ''
    cli.prompt_session = cast(Any, SimpleNamespace(app=None))
    assert to_plain_text(repl_mode.get_custom_toolbar(cli, 'fmt')) == ''

    cli.prompt_session = cast(Any, FakePromptSession())
    cli.last_custom_toolbar_message = repl_mode.ANSI('cached')
    cli.prompt_session.app.current_buffer.text = 'typing'
    assert repl_mode.get_custom_toolbar(cli, 'fmt') == cli.last_custom_toolbar_message

    cli.prompt_session.app.current_buffer.text = ''
    monkeypatch.setattr(repl_mode, 'render_prompt_string', lambda mycli, string, render_counter: f'title:{string}')
    assert 'title:fmt' in str(repl_mode.get_custom_toolbar(cli, 'fmt'))

    cli.terminal_tab_title_format = 'tab'
    cli.terminal_window_title_format = 'window'
    cli.multiplex_window_title_format = 'mux-window'
    cli.multiplex_pane_title_format = 'mux-pane'
    monkeypatch.setattr(repl_mode, 'sanitize_terminal_title', lambda title: title.upper())
    monkeypatch.setattr(repl_mode.sys.stderr, 'isatty', lambda: True)
    printed: list[str] = []
    monkeypatch.setattr(builtins, 'print', lambda *args, **kwargs: printed.append(args[0]))
    tmux_calls: list[tuple[Any, ...]] = []
    monkeypatch.setattr(repl_mode.subprocess, 'run', lambda *args, **kwargs: tmux_calls.append(args))
    monkeypatch.setenv('TMUX', '1')
    repl_mode.set_all_external_titles(cli)
    assert printed[0].startswith('\x1b]1;TITLE:TAB')
    assert printed[1].startswith('\x1b]2;TITLE:WINDOW')
    assert printed[2].startswith('\x1b]2;TITLE:MUX-PANE')
    assert tmux_calls

    monkeypatch.setattr(repl_mode.sys.stderr, 'isatty', lambda: False)
    repl_mode.set_external_terminal_tab_title(cli)
    repl_mode.set_external_terminal_window_title(cli)
    repl_mode.set_external_multiplex_pane_title(cli)
    monkeypatch.delenv('TMUX', raising=False)
    repl_mode.set_external_multiplex_window_title(cli)
    monkeypatch.setenv('TMUX', '1')
    monkeypatch.setattr(repl_mode.subprocess, 'run', lambda *args, **kwargs: (_ for _ in ()).throw(FileNotFoundError()))
    repl_mode.set_external_multiplex_window_title(cli)


def test_prompt_and_title_helper_early_returns_and_remaining_prompt_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    class PromptCursor:
        def __enter__(self) -> 'PromptCursor':
            return self

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> Literal[False]:
            return False

    class PromptConnection:
        def cursor(self) -> PromptCursor:
            return PromptCursor()

    cli = make_repl_cli(
        SimpleNamespace(
            user='alice',
            host=None,
            dbname='db',
            port=3306,
            socket=None,
            server_info=SimpleNamespace(species=SimpleNamespace(name='MySQL')),
            conn=PromptConnection(),
        )
    )
    cli.prompt_session = cast(Any, FakePromptSession())

    monkeypatch.setattr(repl_mode, 'get_uptime', lambda cur: 123)
    monkeypatch.setattr(repl_mode, 'format_uptime', lambda uptime: f'uptime:{uptime}')
    monkeypatch.setattr(repl_mode, 'get_ssl_version', lambda cur: 'TLSv1.3')
    monkeypatch.setattr(repl_mode, 'get_warning_count', lambda cur: 7)

    prompt = repl_mode.render_prompt_string(cli, r'\h|\H|\y|\Y', 0)
    prompt_plain = to_plain_text(prompt)
    assert prompt_plain == f'{repl_mode.DEFAULT_HOST}|{repl_mode.DEFAULT_HOST}|123|uptime:123'

    prompt = repl_mode.render_prompt_string(cli, r'\h|\H|\w|\W', 1)
    prompt_plain = to_plain_text(prompt)
    assert prompt_plain == f'{repl_mode.DEFAULT_HOST}|{repl_mode.DEFAULT_HOST}|7|7'

    prompt = repl_mode.render_prompt_string(cli, r'\h|\H|\T', 2)
    prompt_plain = to_plain_text(prompt)
    assert prompt_plain == f'{repl_mode.DEFAULT_HOST}|{repl_mode.DEFAULT_HOST}|TLSv1.3'

    monkeypatch.setattr(repl_mode.sys.stderr, 'isatty', lambda: True)
    monkeypatch.setattr(builtins, 'print', lambda *args, **kwargs: (_ for _ in ()).throw(AssertionError('unexpected print')))
    monkeypatch.setattr(
        repl_mode.subprocess,
        'run',
        lambda *args, **kwargs: (_ for _ in ()).throw(AssertionError('unexpected tmux call')),
    )

    cli.terminal_tab_title_format = ''
    repl_mode.set_external_terminal_tab_title(cli)
    cli.terminal_tab_title_format = 'tab'
    cli.prompt_session = None
    repl_mode.set_external_terminal_tab_title(cli)

    cli.prompt_session = cast(Any, FakePromptSession())
    cli.terminal_window_title_format = ''
    repl_mode.set_external_terminal_window_title(cli)
    cli.terminal_window_title_format = 'window'
    cli.prompt_session = None
    repl_mode.set_external_terminal_window_title(cli)

    cli.prompt_session = cast(Any, FakePromptSession())
    cli.multiplex_window_title_format = ''
    repl_mode.set_external_multiplex_window_title(cli)
    cli.multiplex_window_title_format = 'mux-window'
    monkeypatch.setenv('TMUX', '1')
    cli.prompt_session = None
    repl_mode.set_external_multiplex_window_title(cli)

    cli.prompt_session = cast(Any, FakePromptSession())
    cli.multiplex_pane_title_format = ''
    repl_mode.set_external_multiplex_pane_title(cli)
    cli.multiplex_pane_title_format = 'mux-pane'
    monkeypatch.delenv('TMUX', raising=False)
    repl_mode.set_external_multiplex_pane_title(cli)
    monkeypatch.setenv('TMUX', '1')
    cli.prompt_session = None
    repl_mode.set_external_multiplex_pane_title(cli)


def test_maybe_html_escape() -> None:
    assert repl_mode.maybe_html_escape('plain', False) == 'plain'
    assert repl_mode.maybe_html_escape('a&b<1>', True) == 'a&amp;b&lt;1&gt;'


def test_render_prompt_string_html() -> None:
    repl_mode.render_prompt_string.cache_clear()

    cli = make_repl_cli(
        SimpleNamespace(
            user='ab',
            host='db.example.com',
            dbname='nameprod',
            port=3306,
            socket=None,
            server_info=SimpleNamespace(species=SimpleNamespace(name='MySQL')),
            conn=None,
        )
    )
    cli.dsn_alias = 'aliasone'

    html_prompt = repl_mode.render_prompt_string(cli, r'\<html><b>\u@\d|\A</b>\</html>', 0)
    assert to_plain_text(html_prompt) == 'ab@nameprod|aliasone'

    bad_html_prompt = repl_mode.render_prompt_string(cli, r'\<html><b>\u', 1)
    assert to_plain_text(bad_html_prompt) == '(cannot parse HTML prompt string)'

    ansi_prompt = repl_mode.render_prompt_string(cli, r'\x1b[31mred\x1b[0m', 2)
    assert to_plain_text(ansi_prompt) == 'red'


def test_render_prompt_string_ansi() -> None:
    repl_mode.render_prompt_string.cache_clear()

    cli = make_repl_cli(
        SimpleNamespace(
            user='ab',
            host='db.example.com',
            dbname='nameprod',
            port=3306,
            socket=None,
            server_info=SimpleNamespace(species=SimpleNamespace(name='MySQL')),
            conn=None,
        )
    )
    cli.dsn_alias = 'aliasone'

    ansi_prompt = repl_mode.render_prompt_string(cli, r'\x1b[31mred\x1b[0m', 2)
    assert to_plain_text(ansi_prompt) == 'red'


def test_output_results_covers_watch_warning_timing_beep_and_interrupts(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeSQLExecute:
        def run(self, text: str) -> list[SQLResult]:
            assert text == 'SHOW WARNINGS'
            return [SQLResult(status='warning', rows=[('warn',)])]

    cli = make_repl_cli(FakeSQLExecute())
    cli.auto_vertical_output = True
    cli.prompt_session = FakePromptSession(columns=91)
    cli.beep_after_seconds = 0.1
    state = repl_mode.ReplState()
    format_widths: list[int | None] = []

    def format_sqlresult(result: SQLResult, **kwargs: Any) -> Iterator[str]:
        format_widths.append(kwargs.get('max_width'))
        return iter([result.status_plain or 'row'])

    cli.format_sqlresult = format_sqlresult
    time_values = iter([0.2, 1.0, 2.0, 3.0, 3.2])
    monkeypatch.setattr(repl_mode.time, 'time', lambda: next(time_values))
    monkeypatch.setattr(repl_mode.special, 'is_expanded_output', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'is_show_warnings_enabled', lambda: True)
    monkeypatch.setattr(repl_mode.special, 'is_timing_enabled', lambda: True)
    monkeypatch.setattr(repl_mode, 'Cursor', FakeCursorBase)
    monkeypatch.setattr(repl_mode, 'is_select', lambda status: False)
    monkeypatch.setattr(repl_mode, 'is_mutating', lambda status: status == 'mut')

    results = sqlresult_generator(
        SQLResult(status='watch', command={'name': 'watch', 'seconds': '1'}),
        SQLResult(status='mut', rows=cast(Any, FakeCursorBase(rowcount=1, warning_count=1))),
    )

    repl_mode._output_results(cli, state, results, start=0.0)

    assert state.mutating is True
    assert format_widths[:2] == [91, 91]
    assert cli.prompt_session.output.bell_count == 2
    assert '' in cli.echo_calls
    assert any(is_warnings_style is True for _, _, is_warnings_style in cli.output_calls)
    assert any(is_warnings_style is False for _, is_warnings_style in cli.timing_calls)
    assert any(is_warnings_style is True for _, is_warnings_style in cli.timing_calls)

    cli_interrupt = make_repl_cli(SimpleNamespace())
    cli_interrupt.echo = lambda message, **kwargs: (
        (_ for _ in ()).throw(KeyboardInterrupt()) if message == '' else cli_interrupt.echo_calls.append(str(message))
    )
    cli_interrupt.output = lambda formatted, result, is_warnings_style=False: (_ for _ in ()).throw(KeyboardInterrupt())
    monkeypatch.setattr(repl_mode.special, 'is_timing_enabled', lambda: False)
    monkeypatch.setattr(repl_mode, 'is_select', lambda status: False)
    monkeypatch.setattr(repl_mode.time, 'time', lambda: 0.0)
    repl_mode._output_results(
        cli_interrupt,
        repl_mode.ReplState(),
        sqlresult_generator(SQLResult(status='first'), SQLResult(status='second')),
        start=0.0,
    )


def test_output_results_handles_abort_default_width_and_bad_watch(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_repl_cli(SimpleNamespace())
    cli.auto_vertical_output = True
    widths: list[int | None] = []

    def format_sqlresult_with_width(result: SQLResult, **kwargs: Any) -> Iterator[str]:
        widths.append(kwargs.get('max_width'))
        return iter([result.status_plain or 'row'])

    cli.format_sqlresult = format_sqlresult_with_width
    monkeypatch.setattr(repl_mode, 'Cursor', FakeCursorBase)
    monkeypatch.setattr(repl_mode, 'is_select', lambda status: status == 'select')
    monkeypatch.setattr(repl_mode, 'confirm', lambda text: False)
    repl_mode._output_results(
        cli,
        repl_mode.ReplState(),
        sqlresult_generator(SQLResult(status='select', rows=cast(Any, FakeCursorBase(rowcount=1001)))),
        start=0.0,
    )
    assert 'The result set has more than 1000 rows.' in cli.echo_calls
    assert 'Aborted!' in cli.echo_calls

    repl_mode._output_results(
        cli,
        repl_mode.ReplState(),
        sqlresult_generator(SQLResult(status='ok')),
        start=0.0,
    )
    assert widths[-1] == repl_mode.DEFAULT_WIDTH

    monkeypatch.setattr(repl_mode, 'is_select', lambda status: False)
    with pytest.raises(SystemExit):
        repl_mode._output_results(
            cli,
            repl_mode.ReplState(),
            sqlresult_generator(
                SQLResult(status='watch', command={'name': 'watch', 'seconds': '1'}),
                SQLResult(status='watch', command={'name': 'watch', 'seconds': 'bad'}),
            ),
            start=0.0,
        )


def test_keepalive_hook_covers_threshold_and_errors() -> None:
    cli = make_repl_cli(SimpleNamespace(conn=FakeConnection()))
    repl_mode._keepalive_hook(cli, None)
    assert cli._keepalive_counter == 0

    cli.keepalive_ticks = 0
    repl_mode._keepalive_hook(cli, None)
    assert cli._keepalive_counter == 0

    cli.keepalive_ticks = 1
    repl_mode._keepalive_hook(cli, None)
    assert cli._keepalive_counter == 1
    repl_mode._keepalive_hook(cli, None)
    assert cli._keepalive_counter == 0
    assert cli.sqlexecute.conn.ping_calls == [False]

    cli.sqlexecute.conn = FakeConnection(ping_exc=RuntimeError('boom'))
    repl_mode._keepalive_hook(cli, None)
    repl_mode._keepalive_hook(cli, None)
    assert any('keepalive ping error' in call[0][0] for call in cli.logger.debug_calls)


def test_build_prompt_session_covers_toolbar_modes_and_editing_modes(monkeypatch: pytest.MonkeyPatch) -> None:
    captured_kwargs: list[dict[str, Any]] = []
    toolbar_help: list[bool] = []

    def fake_prompt_session(**kwargs: Any) -> FakePromptSession:
        captured_kwargs.append(kwargs)
        return FakePromptSession()

    monkeypatch.setattr(repl_mode, 'PromptSession', fake_prompt_session)
    monkeypatch.setattr(repl_mode, 'style_factory_ptoolkit', lambda *args, **kwargs: 'style')
    monkeypatch.setattr(repl_mode, 'cli_is_multiline', lambda mycli: False)

    def fake_toolbar_tokens(mycli: Any, show_help: Any, fmt: str, custom_toolbar: Any) -> str:
        toolbar_help.append(show_help())
        assert callable(custom_toolbar)
        return 'toolbar'

    monkeypatch.setattr(repl_mode, 'create_toolbar_tokens_func', fake_toolbar_tokens)

    cli = make_repl_cli(SimpleNamespace())
    state = repl_mode.ReplState()
    cli.toolbar_format = 'none'
    cli.key_bindings = 'vi'
    cli.wider_completion_menu = True
    repl_mode._build_prompt_session(cli, state, history=cast(Any, 'history'), key_bindings=cast(Any, 'bindings'))
    first_kwargs = captured_kwargs[-1]
    assert first_kwargs['bottom_toolbar'] is None
    assert first_kwargs['complete_style'] == repl_mode.CompleteStyle.MULTI_COLUMN
    assert first_kwargs['editing_mode'] == repl_mode.EditingMode.VI
    assert cli.prompt_session.app.ttimeoutlen == cli.vi_ttimeoutlen

    cli.toolbar_format = 'default'
    cli.key_bindings = 'emacs'
    cli.wider_completion_menu = False
    state.iterations = 0
    repl_mode._build_prompt_session(cli, state, history=cast(Any, 'history'), key_bindings=cast(Any, 'bindings'))
    latest_kwargs = captured_kwargs[-1]
    assert latest_kwargs['bottom_toolbar'] == 'toolbar'
    assert latest_kwargs['complete_style'] == repl_mode.CompleteStyle.COLUMN
    assert latest_kwargs['editing_mode'] == repl_mode.EditingMode.EMACS
    assert toolbar_help == [True]
    assert cli.prompt_session.app.ttimeoutlen == cli.emacs_ttimeoutlen
    assert latest_kwargs['prompt_continuation'](4, 0, 0) == [('class:continuation', '  > ')]


def test_one_iteration_handles_prompt_interrupt_empty_editor_clip_and_clip_true(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)
    cli = make_repl_cli(SimpleNamespace(run=lambda text: iter([SQLResult(status='ok')]), conn=FakeConnection()))
    cli.keepalive_ticks = 1
    cli.prompt_session = FakePromptSession([KeyboardInterrupt(), '   ', 'edit-error', 'clip-error', 'clip-stop'])

    repl_mode._one_iteration(cli, repl_mode.ReplState())
    assert cli.query_history == []

    repl_mode._one_iteration(cli, repl_mode.ReplState())
    assert cli.query_history == []
    inputhook = cli.prompt_session.prompt_calls[-1]['inputhook']
    assert inputhook is not None
    inputhook(None)

    monkeypatch.setattr(repl_mode, 'handle_editor_command', lambda *args: (_ for _ in ()).throw(RuntimeError('edit boom')))
    repl_mode._one_iteration(cli, repl_mode.ReplState())
    assert 'edit boom' in cli.echo_calls[-1]

    monkeypatch.setattr(repl_mode, 'handle_editor_command', lambda mycli, text, inputhook, loaded_message_fn: text)
    monkeypatch.setattr(repl_mode, 'handle_clip_command', lambda mycli, text: (_ for _ in ()).throw(RuntimeError('clip boom')))
    repl_mode._one_iteration(cli, repl_mode.ReplState())
    assert 'clip boom' in cli.echo_calls[-1]

    monkeypatch.setattr(repl_mode, 'handle_clip_command', lambda mycli, text: True)
    repl_mode._one_iteration(cli, repl_mode.ReplState())
    assert cli.query_history == []


def test_one_iteration_covers_llm_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)
    click_output: list[str] = []
    monkeypatch.setattr(repl_mode.click, 'echo', lambda message='', **kwargs: click_output.append(str(message)))
    monkeypatch.setattr(repl_mode.special, 'is_timing_enabled', lambda: True)
    monkeypatch.setattr(repl_mode.special, 'is_llm_command', lambda text: text.startswith('\\llm'))

    class FakeSQLExecute:
        def __init__(self) -> None:
            self.dbname = 'db'
            self.conn = FakeConnection(cursor_value='cursor')

        def run(self, text: str) -> Iterator[SQLResult]:
            return iter([SQLResult(status=f'ran:{text}')])

    monkeypatch.setattr(
        repl_mode.special,
        'handle_llm',
        lambda text, cur, dbname, field_truncate, section_truncate: ('context', 'select 1', 1.25),
    )
    cli = make_repl_cli(FakeSQLExecute())
    cli.prompt_session = FakePromptSession(['\\llm ask', 'select 1'])
    repl_mode._one_iteration(
        cli,
        repl_mode.ReplState(),
    )
    assert click_output[:3] == ['LLM Response:', 'context', '---']
    assert cli.output_calls[0][0] == ['None', 'ran:select 1']

    cli_finish = make_repl_cli(FakeSQLExecute())
    cli_finish.prompt_session = FakePromptSession(['\\llm finish'])
    cli_finish.format_sqlresult = lambda result, **kwargs: iter([result.status_plain or 'row'])
    monkeypatch.setattr(
        repl_mode.special,
        'handle_llm',
        lambda *args, **kwargs: (_ for _ in ()).throw(repl_mode.special.FinishIteration(iter([SQLResult(status='done')]))),
    )
    repl_mode._one_iteration(cli_finish, repl_mode.ReplState())
    assert cli_finish.output_calls[0][0] == ['done']

    cli_empty = make_repl_cli(FakeSQLExecute())
    cli_empty.prompt_session = FakePromptSession(['\\llm empty'])
    monkeypatch.setattr(
        repl_mode.special,
        'handle_llm',
        lambda *args, **kwargs: (_ for _ in ()).throw(repl_mode.special.FinishIteration(None)),
    )
    repl_mode._one_iteration(cli_empty, repl_mode.ReplState())
    assert cli_empty.output_calls == []

    cli_err = make_repl_cli(FakeSQLExecute())
    cli_err.prompt_session = FakePromptSession(['\\llm err'])
    monkeypatch.setattr(
        repl_mode.special,
        'handle_llm',
        lambda *args, **kwargs: (_ for _ in ()).throw(RuntimeError('llm boom')),
    )
    repl_mode._one_iteration(cli_err, repl_mode.ReplState())
    assert 'llm boom' in cli_err.echo_calls[-1]

    cli_interrupt = make_repl_cli(FakeSQLExecute())
    cli_interrupt.prompt_session = FakePromptSession(['\\llm stop'])
    monkeypatch.setattr(
        repl_mode.special,
        'handle_llm',
        lambda *args, **kwargs: (_ for _ in ()).throw(KeyboardInterrupt()),
    )
    repl_mode._one_iteration(cli_interrupt, repl_mode.ReplState())
    assert cli_interrupt.output_calls == []

    cli_quiet = make_repl_cli(FakeSQLExecute())
    cli_quiet.prompt_session = FakePromptSession(['\\llm quiet', 'select 2'])
    monkeypatch.setattr(repl_mode.special, 'is_timing_enabled', lambda: False)
    monkeypatch.setattr(
        repl_mode.special,
        'handle_llm',
        lambda text, cur, dbname, field_truncate, section_truncate: ('', 'select 2', 0.5),
    )
    repl_mode._one_iteration(cli_quiet, repl_mode.ReplState())
    assert cli_quiet.output_calls[0][0] == ['None', 'ran:select 2']


@pytest.mark.parametrize(
    'text, expected',
    [
        ('', True),
        ('  ', True),
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'new'", True),
        ('alter user root identified by "pw"', True),
        ("SET PASSWORD = 'newpass'", True),
        ("set password = 'newpass'", True),
        ('quit', True),
        ('exit', True),
        ('\\q', True),
        ('SELECT 1', False),
        ('DROP TABLE t', False),
        ('USE mydb', False),
        ('SHOW DATABASES', False),
    ],
)
def test_is_sandbox_allowed(text: str, expected: bool) -> None:
    from mycli.packages.sql_utils import is_sandbox_allowed

    assert is_sandbox_allowed(text) is expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'new'", True),
        ("SET PASSWORD = 'newpass'", True),
        ('SELECT 1', False),
        ('quit', False),
    ],
)
def test_is_password_change(text: str, expected: bool) -> None:
    from mycli.packages.sql_utils import is_password_change

    assert is_password_change(text) is expected


@pytest.mark.parametrize(
    'text, expected',
    [
        ("ALTER USER 'root'@'localhost' IDENTIFIED BY 'newpass'", 'newpass'),
        ("SET PASSWORD = 'secret123'", 'secret123'),
        ("ALTER USER root IDENTIFIED BY 'p@ss w0rd!'", 'p@ss w0rd!'),
        ('ALTER USER root IDENTIFIED WITH mysql_native_password', None),
        ('SELECT 1', None),
    ],
)
def test_extract_new_password(text: str, expected: str | None) -> None:
    from mycli.packages.sql_utils import extract_new_password

    assert extract_new_password(text) == expected


def test_one_iteration_blocks_disallowed_in_sandbox_mode(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class FakeSQLExecute:
        def __init__(self) -> None:
            self.dbname = 'db'
            self.connection_id = 0

        def run(self, text: str) -> Iterator[SQLResult]:
            return iter([SQLResult(status=f'ran:{text}')])

    cli = make_repl_cli(FakeSQLExecute())
    cli.sandbox_mode = True

    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'SELECT 1')
    assert any('ERROR 1820' in msg for msg in cli.echo_calls)
    assert not cli.query_history


def test_one_iteration_allows_alter_user_in_sandbox_mode(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class FakeSQLExecute:
        def __init__(self) -> None:
            self.dbname = 'db'
            self.connection_id = 0
            self.password = 'old'
            self.connect_calls: list[bool] = []

        def connect(self) -> None:
            self.connect_calls.append(True)

        def run(self, text: str) -> Iterator[SQLResult]:
            return iter([SQLResult(status='OK')])

    sqlexecute = FakeSQLExecute()
    cli = make_repl_cli(sqlexecute)
    cli.sandbox_mode = True
    monkeypatch.setattr(repl_mode, 'is_mutating', lambda status: False)

    repl_mode._one_iteration(cli, repl_mode.ReplState(), "ALTER USER 'root'@'localhost' IDENTIFIED BY 'newpass'")
    assert cli.sandbox_mode is False
    assert sqlexecute.password == 'newpass'
    assert sqlexecute.connect_calls == [True]
    assert any('Reconnected' in msg for msg in cli.echo_calls)


def test_one_iteration_sandbox_reconnect_failure(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class FakeSQLExecute:
        def __init__(self) -> None:
            self.dbname = 'db'
            self.connection_id = 0
            self.password = 'old'

        def connect(self) -> None:
            raise RuntimeError('connection refused')

        def run(self, text: str) -> Iterator[SQLResult]:
            return iter([SQLResult(status='OK')])

    sqlexecute = FakeSQLExecute()
    cli = make_repl_cli(sqlexecute)
    cli.sandbox_mode = True
    monkeypatch.setattr(repl_mode, 'is_mutating', lambda status: False)

    repl_mode._one_iteration(cli, repl_mode.ReplState(), "ALTER USER 'root'@'localhost' IDENTIFIED BY 'newpass'")
    assert cli.sandbox_mode is False
    assert any('reconnection failed' in msg for msg in cli.echo_calls)


def test_one_iteration_enters_sandbox_mode_on_must_change_password_error(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class FakeSQLExecute:
        dbname = 'db'
        connection_id = 0

        def run(self, text: str) -> Iterator[SQLResult]:
            raise pymysql.OperationalError(repl_mode.ER_MUST_CHANGE_PASSWORD, 'must change password')

    cli = make_repl_cli(FakeSQLExecute())

    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'SELECT 1')

    assert cli.sandbox_mode is True
    assert any('ERROR 1820' in msg for msg in cli.echo_calls)
    assert cli.query_history[-1].query == 'SELECT 1'
    assert cli.query_history[-1].successful is False


def test_one_iteration_covers_redirect_destructive_success_refresh_and_logfile(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class FakeSQLExecute:
        def __init__(self) -> None:
            self.dbname: str | None = 'db'
            self.connection_id = 0
            self.calls: list[str] = []

        def connect(self) -> None:
            self.calls.append('connect')

        def run(self, text: str) -> Iterator[SQLResult]:
            self.calls.append(text)
            return iter([SQLResult(status='DROP 1')])

    sqlexecute = FakeSQLExecute()
    cli = make_repl_cli(sqlexecute)
    cli.logfile = False
    cli.destructive_warning = True
    monkeypatch.setattr(repl_mode, 'is_redirect_command', lambda text: text == 'redirect')
    monkeypatch.setattr(repl_mode, 'get_redirect_components', lambda text: ('dropdb', 'tee', '>', 'out.txt'))
    redirects: list[tuple[Any, ...]] = []
    monkeypatch.setattr(repl_mode.special, 'set_redirect', lambda *args: redirects.append(args))
    monkeypatch.setattr(
        repl_mode,
        'confirm_destructive_query',
        lambda keywords, text: None if text == 'dropdb' else (True if text == 'approved' else False),
    )
    monkeypatch.setattr(repl_mode, 'is_dropping_database', lambda text, dbname: text == 'dropdb')
    monkeypatch.setattr(repl_mode, 'need_completion_refresh', lambda text: text == 'dropdb')
    monkeypatch.setattr(repl_mode, 'need_completion_reset', lambda text: text == 'dropdb')
    monkeypatch.setattr(repl_mode, 'is_mutating', lambda status: True)

    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'redirect')
    assert redirects == [('tee', '>', 'out.txt')]
    assert cli.refresh_calls == [True]
    assert cli.query_history[-1].query == 'dropdb'
    assert cli.query_history[-1].successful is True
    assert cli.query_history[-1].mutating is True
    assert sqlexecute.dbname is None
    assert sqlexecute.calls == ['dropdb', 'connect']
    assert 'Warning: This query was not logged.' in cli.echo_calls

    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'approved')
    assert 'Your call!' in cli.echo_calls

    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'denied')
    assert 'Wise choice!' in cli.echo_calls


def test_one_iteration_covers_reconnect_and_error_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class InterfaceSQLExecute:
        def __init__(self) -> None:
            self.dbname: str | None = 'db'
            self.connection_id = 0
            self.calls: list[str] = []

        def run(self, text: str) -> Iterator[SQLResult]:
            self.calls.append(text)
            if text == 'iface' and self.calls.count('iface') == 1:
                raise pymysql.err.InterfaceError()
            return iter([SQLResult(status=f'ok:{text}')])

    interface_sql = InterfaceSQLExecute()
    cli_interface = make_repl_cli(interface_sql)
    interface_reconnect_calls: list[str] = []
    interface_results = iter([True])

    def interface_reconnect(database: str = '') -> bool:
        interface_reconnect_calls.append(database)
        return next(interface_results)

    cli_interface.reconnect = interface_reconnect

    repl_mode._one_iteration(cli_interface, repl_mode.ReplState(), 'iface')
    assert interface_sql.calls.count('iface') == 2
    assert cli_interface.query_history[-1].query == 'iface'
    assert interface_reconnect_calls == ['']

    cli_interface_false = make_repl_cli(InterfaceSQLExecute())
    false_calls: list[str] = []

    def interface_reconnect_false(database: str = '') -> bool:
        false_calls.append(database)
        return False

    cli_interface_false.reconnect = interface_reconnect_false
    repl_mode._one_iteration(cli_interface_false, repl_mode.ReplState(), 'iface')
    assert false_calls == ['']

    class ErrorSQLExecute:
        def __init__(self) -> None:
            self.dbname: str | None = 'db'
            self.connection_id = 0
            self.calls: list[str] = []

        def run(self, text: str) -> Iterator[SQLResult]:
            self.calls.append(text)
            if text == 'oplost' and self.calls.count('oplost') == 1:
                raise pymysql.OperationalError(2003, 'lost')
            if text == 'opbad':
                raise pymysql.OperationalError(9999, 'bad op')
            if text == 'nyi':
                raise NotImplementedError()
            if text == 'boom':
                raise RuntimeError('boom')
            if text == 'eof':
                raise EOFError()
            return iter([SQLResult(status=f'ok:{text}')])

    error_sql = ErrorSQLExecute()
    cli_error = make_repl_cli(error_sql)
    error_reconnect_calls: list[str] = []

    def error_reconnect(database: str = '') -> bool:
        error_reconnect_calls.append(database)
        return True

    cli_error.reconnect = error_reconnect

    repl_mode._one_iteration(cli_error, repl_mode.ReplState(), 'oplost')
    assert error_sql.calls.count('oplost') == 2
    repl_mode._one_iteration(cli_error, repl_mode.ReplState(), 'opbad')
    repl_mode._one_iteration(cli_error, repl_mode.ReplState(), 'nyi')
    repl_mode._one_iteration(cli_error, repl_mode.ReplState(), 'boom')
    assert any('bad op' in line for line in cli_error.echo_calls)
    assert 'Not Yet Implemented.' in cli_error.echo_calls
    assert any('boom' in line for line in cli_error.echo_calls)
    assert error_reconnect_calls == ['']

    cli_error_false = make_repl_cli(ErrorSQLExecute())
    false_reconnect_calls: list[str] = []

    def error_reconnect_false(database: str = '') -> bool:
        false_reconnect_calls.append(database)
        return False

    cli_error_false.reconnect = error_reconnect_false
    repl_mode._one_iteration(cli_error_false, repl_mode.ReplState(), 'oplost')
    assert false_reconnect_calls == ['']

    with pytest.raises(EOFError):
        repl_mode._one_iteration(cli_error, repl_mode.ReplState(), 'eof')


def test_one_iteration_reraises_eoferror(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class EofSQLExecute:
        dbname = 'db'
        connection_id = 0

        def run(self, text: str) -> Iterator[SQLResult]:
            raise EOFError()

    with pytest.raises(EOFError):
        repl_mode._one_iteration(make_repl_cli(EofSQLExecute()), repl_mode.ReplState(), 'eof')


def test_one_iteration_covers_cancel_paths_and_redirect_error(monkeypatch: pytest.MonkeyPatch) -> None:
    patch_repl_runtime_defaults(monkeypatch)

    class FakeSQLExecute:
        def __init__(self) -> None:
            self.dbname = 'db'
            self.connection_id = 0

        def connect(self) -> None:
            return None

        def run(self, text: str) -> Iterator[SQLResult]:
            if text == 'cancel-ok':
                self.connection_id = 7
                raise KeyboardInterrupt()
            if text == 'kill 7':
                return iter([SQLResult(status='OK')])
            if text == 'cancel-fail':
                self.connection_id = 8
                raise KeyboardInterrupt()
            if text == 'kill 8':
                return iter([SQLResult(status='failed')])
            if text == 'cancel-error':
                self.connection_id = 9
                raise KeyboardInterrupt()
            if text == 'kill 9':
                raise RuntimeError('kill failed')
            if text == 'cancel-missing':
                self.connection_id = 0
                raise KeyboardInterrupt()
            return iter([SQLResult(status='ok')])

    cli = make_repl_cli(FakeSQLExecute())
    monkeypatch.setattr(repl_mode, 'is_redirect_command', lambda text: text == 'redirect-bad')
    monkeypatch.setattr(repl_mode, 'get_redirect_components', lambda text: ('sql', 'tee', '>', 'out.txt'))
    monkeypatch.setattr(repl_mode.special, 'set_redirect', lambda *args: (_ for _ in ()).throw(RuntimeError('redirect boom')))
    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'redirect-bad')
    assert 'redirect boom' in cli.echo_calls[-1]

    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'cancel-ok')
    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'cancel-fail')
    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'cancel-error')
    repl_mode._one_iteration(cli, repl_mode.ReplState(), 'cancel-missing')
    assert 'Cancelled query id: 7' in cli.echo_calls
    assert any('Failed to confirm query cancellation' in line for line in cli.echo_calls)
    assert any('Encountered error while cancelling query' in line for line in cli.echo_calls)
    assert 'Did not get a connection id, skip cancelling query' in cli.echo_calls


def test_main_repl_covers_setup_loop_and_goodbye(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_repl_cli(SimpleNamespace())
    cli.verbosity = 0
    cli.smart_completion = True
    loop_iterations: list[int] = []
    monkeypatch.setattr(repl_mode, '_create_history', lambda mycli: 'history')
    monkeypatch.setattr(repl_mode, 'mycli_bindings', lambda mycli: 'bindings')
    monkeypatch.setattr(repl_mode, '_show_startup_banner', lambda mycli, sqlexecute: None)
    monkeypatch.setattr(
        repl_mode,
        '_build_prompt_session',
        lambda mycli, state, history, key_bindings: setattr(mycli, 'prompt_session', FakePromptSession()),
    )

    def fake_one_iteration(mycli: Any, state: repl_mode.ReplState) -> None:
        loop_iterations.append(state.iterations)
        if len(loop_iterations) == 2:
            raise EOFError()

    closed: list[bool] = []
    monkeypatch.setattr(repl_mode, '_one_iteration', fake_one_iteration)
    monkeypatch.setattr(repl_mode.special, 'close_tee', lambda: closed.append(True))
    monkeypatch.setattr(repl_mode, 'set_all_external_titles', lambda mycli: setattr(mycli, 'title_calls', mycli.title_calls + 1))

    repl_mode.main_repl(cli)

    assert cli.pager_configured == 1
    assert cli.refresh_calls == [False]
    assert cli.title_calls == 1
    assert loop_iterations == [0, 1]
    assert closed == [True]
    assert cli.echo_calls[-1] == 'Goodbye!'


def test_main_repl_covers_no_refresh_and_quiet_exit(monkeypatch: pytest.MonkeyPatch) -> None:
    cli = make_repl_cli(SimpleNamespace())
    cli.verbosity = -1
    cli.smart_completion = False
    monkeypatch.setattr(repl_mode, '_create_history', lambda mycli: 'history')
    monkeypatch.setattr(repl_mode, 'mycli_bindings', lambda mycli: 'bindings')
    monkeypatch.setattr(repl_mode, '_show_startup_banner', lambda mycli, sqlexecute: None)
    monkeypatch.setattr(
        repl_mode,
        '_build_prompt_session',
        lambda mycli, state, history, key_bindings: setattr(mycli, 'prompt_session', FakePromptSession()),
    )
    monkeypatch.setattr(repl_mode, '_one_iteration', lambda mycli, state: (_ for _ in ()).throw(EOFError()))
    monkeypatch.setattr(repl_mode.special, 'close_tee', lambda: None)
    monkeypatch.setattr(repl_mode, 'set_all_external_titles', lambda mycli: setattr(mycli, 'title_calls', mycli.title_calls + 1))

    repl_mode.main_repl(cli)

    assert cli.refresh_calls == []
    assert cli.echo_calls == []


def test_output_results_covers_remaining_watch_select_and_warning_branches(monkeypatch: pytest.MonkeyPatch) -> None:
    class WarninglessSQLExecute:
        def run(self, text: str) -> list[SQLResult]:
            assert text == 'SHOW WARNINGS'
            return []

    cli = make_repl_cli(WarninglessSQLExecute())
    cli.show_warnings = True
    cli.auto_vertical_output = False
    cli.prompt_session = FakePromptSession(columns=77)
    monkeypatch.setattr(repl_mode, 'Cursor', FakeCursorBase)
    monkeypatch.setattr(repl_mode, 'is_mutating', lambda status: False)
    monkeypatch.setattr(repl_mode, 'confirm', lambda text: True)
    monkeypatch.setattr(repl_mode.special, 'is_expanded_output', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'is_redirected', lambda: False)
    monkeypatch.setattr(repl_mode.special, 'is_timing_enabled', lambda: True)
    monkeypatch.setattr(repl_mode, 'is_select', lambda status: status == 'select')
    monkeypatch.setattr(repl_mode.time, 'time', lambda: 0.0)

    repl_mode._output_results(
        cli,
        repl_mode.ReplState(),
        sqlresult_generator(
            SQLResult(status='watch', command={'name': 'watch', 'seconds': '1'}),
            SQLResult(status='watch', command={'name': 'watch', 'seconds': '2'}),
            SQLResult(status='select', rows=cast(Any, FakeCursorBase(rowcount=1001, warning_count=1))),
        ),
        start=0.0,
    )
    assert cli.output_calls



from __future__ import annotations

from typing import TYPE_CHECKING

import click

from mycli.packages.ssh_utils import read_ssh_config

if TYPE_CHECKING:
    from mycli.client import MyCli
    from mycli.main import CliArgs


def main_list_ssh_config(mycli: 'MyCli', cli_args: 'CliArgs') -> int:
    ssh_config = read_ssh_config(cli_args.ssh_config_path)
    try:
        host_entries = ssh_config.get_hostnames()
    except KeyError:
        click.secho('Error reading ssh config', err=True, fg="red")
        return 1
    for host_entry in host_entries:
        if mycli.verbosity >= 1:
            host_config = ssh_config.lookup(host_entry)
            click.secho(f"{host_entry} : {host_config.get('hostname')}")
        else:
            click.secho(host_entry)
    return 0



from __future__ import annotations

from dataclasses import dataclass, field
from types import SimpleNamespace
from typing import Any, Callable, cast

import prompt_toolkit
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.controls import BufferControl, SearchBufferControl
from prompt_toolkit.selection import SelectionType
import pytest

from mycli import key_bindings


@dataclass
class DummyKeysConfig:
    behaviors: dict[str, list[str]] = field(default_factory=dict)
    options: dict[str, str] = field(default_factory=dict)

    def as_list(self, name: str) -> list[str]:
        return self.behaviors[name]

    def get(self, name: str, default: str | None = None) -> str | None:
        return self.options.get(name, default)


@dataclass
class DummyOutput:
    bell_calls: int = 0

    def bell(self) -> None:
        self.bell_calls += 1


@dataclass
class DummyBuffer:
    text: str = ''
    complete_state: object | None = None
    complete_next_calls: int = 0
    cancel_completion_calls: int = 0
    open_in_editor_calls: list[bool] = field(default_factory=list)
    start_completion_calls: list[dict[str, bool]] = field(default_factory=list)
    start_selection_calls: list[SelectionType] = field(default_factory=list)
    transform_calls: list[tuple[int, int, Callable[[str], str]]] = field(default_factory=list)
    inserted_text: list[str] = field(default_factory=list)
    validate_calls: int = 0

    def complete_next(self) -> None:
        self.complete_next_calls += 1

    def start_completion(
        self,
        select_first: bool = False,
        insert_common_part: bool = False,
    ) -> None:
        self.start_completion_calls.append({
            'select_first': select_first,
            'insert_common_part': insert_common_part,
        })
        self.complete_state = object()

    def cancel_completion(self) -> None:
        self.cancel_completion_calls += 1
        self.complete_state = None

    def open_in_editor(self, validate_and_handle: bool) -> None:
        self.open_in_editor_calls.append(validate_and_handle)

    def start_selection(self, selection_type: SelectionType) -> None:
        self.start_selection_calls.append(selection_type)

    def transform_region(self, start: int, end: int, handler: Callable[[str], str]) -> None:
        self.transform_calls.append((start, end, handler))

    def insert_text(self, text: str) -> None:
        self.inserted_text.append(text)

    def validate_and_handle(self) -> None:
        self.validate_calls += 1


@dataclass
class DummyApp:
    current_buffer: DummyBuffer
    editing_mode: EditingMode = EditingMode.VI
    ttimeoutlen: float | None = None
    output: DummyOutput = field(default_factory=DummyOutput)
    exit_calls: list[dict[str, Any]] = field(default_factory=list)
    print_calls: list[Any] = field(default_factory=list)

    def exit(self, exception: type[BaseException], style: str) -> None:
        self.exit_calls.append({'exception': exception, 'style': style})

    def print_text(self, text: Any) -> None:
        self.print_calls.append(text)


@dataclass
class DummyMyCli:
    key_config: DummyKeysConfig
    smart_completion: bool = True
    multi_line: bool = False
    key_bindings_mode: str = 'vi'
    highlight_preview: bool = True
    syntax_style: str = 'native'
    emacs_ttimeoutlen: float = 1.5
    vi_ttimeoutlen: float = 0.5
    sqlexecute: object = field(default_factory=object)
    prettify_calls: list[str] = field(default_factory=list)
    unprettify_calls: list[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.completer = SimpleNamespace(smart_completion=self.smart_completion)
        self.key_bindings = self.key_bindings_mode
        self.config = {'keys': self.key_config}

    def handle_prettify_binding(self, text: str) -> str:
        self.prettify_calls.append(text)
        return text

    def handle_unprettify_binding(self, text: str) -> str:
        self.unprettify_calls.append(text)
        return text


def make_event(buffer: DummyBuffer | None = None) -> SimpleNamespace:
    active_buffer = buffer or DummyBuffer()
    app = DummyApp(current_buffer=active_buffer)
    return SimpleNamespace(app=app, current_buffer=active_buffer)


def binding_handler(kb: prompt_toolkit.key_binding.KeyBindings, *keys: str | Keys) -> Callable[[Any], None]:
    expected = tuple(keys)
    for binding in kb.bindings:
        if binding.keys == expected:
            return cast(Callable[[Any], None], binding.handler)
    raise AssertionError(f'binding not found for keys={expected!r}')


def binding_filter(kb: prompt_toolkit.key_binding.KeyBindings, *keys: str | Keys) -> Any:
    expected = tuple(keys)
    for binding in kb.bindings:
        if binding.keys == expected:
            return binding.filter
    raise AssertionError(f'binding not found for keys={expected!r}')


def binding(kb: prompt_toolkit.key_binding.KeyBindings, *keys: str | Keys) -> Any:
    expected = tuple(keys)
    for entry in kb.bindings:
        if entry.keys == expected:
            return entry
    raise AssertionError(f'binding not found for keys={expected!r}')


def patch_filter_app(monkeypatch, app: DummyApp) -> None:
    monkeypatch.setitem(key_bindings.emacs_mode.func.__globals__, 'get_app', lambda: app)
    monkeypatch.setitem(key_bindings.vi_mode.func.__globals__, 'get_app', lambda: app)
    monkeypatch.setitem(key_bindings.completion_is_selected.func.__globals__, 'get_app', lambda: app)
    monkeypatch.setitem(key_bindings.control_is_searchable.func.__globals__, 'get_app', lambda: app)


def test_ctrl_d_condition_depends_on_empty_buffer(monkeypatch) -> None:
    monkeypatch.setattr(key_bindings, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='')))
    assert key_bindings.ctrl_d_condition() is True

    monkeypatch.setattr(key_bindings, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(text='select 1')))
    assert key_bindings.ctrl_d_condition() is False


def test_in_completion_depends_on_complete_state(monkeypatch) -> None:
    monkeypatch.setattr(key_bindings, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(complete_state=object())))
    assert key_bindings.in_completion() is True

    monkeypatch.setattr(key_bindings, 'get_app', lambda: SimpleNamespace(current_buffer=SimpleNamespace(complete_state=None)))
    assert key_bindings.in_completion() is False


def test_print_f1_help_prints_inline_help_and_docs_url(monkeypatch) -> None:
    app = DummyApp(current_buffer=DummyBuffer())
    monkeypatch.setattr(key_bindings, 'get_app', lambda: app)

    key_bindings.print_f1_help()

    assert app.print_calls == [
        '\n',
        [
            ('', 'Inline help — type "'),
            ('bold', 'help'),
            ('', '" or "'),
            ('bold', r'\?'),
            ('', '"\n'),
        ],
        [
            ('', 'Docs index — '),
            ('bold', key_bindings.DOCS_URL),
            ('', '\n'),
        ],
        '\n',
    ]


def test_edit_and_execute_opens_editor_without_validation() -> None:
    event = make_event()

    key_bindings.edit_and_execute(cast(KeyPressEvent, event))

    assert event.current_buffer.open_in_editor_calls == [False]


@pytest.mark.parametrize('keys', ((Keys.F1,), (Keys.Escape, '[', 'P')))
def test_f1_bindings_open_docs_show_help_and_invalidate(monkeypatch, keys: tuple[str | Keys, ...]) -> None:
    mycli = DummyMyCli(DummyKeysConfig())
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event()
    browser_calls: list[str] = []
    terminal_calls: list[Callable[[], None]] = []
    invalidated: list[DummyApp] = []

    monkeypatch.setattr(key_bindings.webbrowser, 'open_new_tab', lambda url: browser_calls.append(url))
    monkeypatch.setattr(
        key_bindings.prompt_toolkit.application,
        'run_in_terminal',
        lambda fn: terminal_calls.append(fn),
    )
    monkeypatch.setattr(key_bindings, 'safe_invalidate_display', lambda app: invalidated.append(app))

    binding_handler(kb, *keys)(event)

    assert browser_calls == [key_bindings.DOCS_URL]
    assert terminal_calls == [key_bindings.print_f1_help]
    assert invalidated == [event.app]


@pytest.mark.parametrize('keys', ((Keys.F2,), (Keys.Escape, '[', 'Q')))
def test_f2_bindings_toggle_smart_completion(keys: tuple[str | Keys, ...]) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), smart_completion=True)
    kb = key_bindings.mycli_bindings(mycli)

    binding_handler(kb, *keys)(make_event())

    assert mycli.completer.smart_completion is False


@pytest.mark.parametrize('keys', ((Keys.F3,), (Keys.Escape, '[', 'R')))
def test_f3_bindings_toggle_multiline_mode(keys: tuple[str | Keys, ...]) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), multi_line=False)
    kb = key_bindings.mycli_bindings(mycli)

    binding_handler(kb, *keys)(make_event())

    assert mycli.multi_line is True


@pytest.mark.parametrize(
    ('keys', 'initial_mode', 'expected_mode', 'expected_editing_mode', 'expected_timeout'),
    (
        ((Keys.F4,), 'vi', 'emacs', EditingMode.EMACS, 1.5),
        ((Keys.F4,), 'emacs', 'vi', EditingMode.VI, 0.5),
        ((Keys.Escape, '[', 'S'), 'vi', 'emacs', EditingMode.EMACS, 1.5),
        ((Keys.Escape, '[', 'S'), 'emacs', 'vi', EditingMode.VI, 0.5),
    ),
)
def test_f4_bindings_toggle_key_binding_modes(
    keys: tuple[str | Keys, ...],
    initial_mode: str,
    expected_mode: str,
    expected_editing_mode: EditingMode,
    expected_timeout: float,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), key_bindings_mode=initial_mode)
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event()

    binding_handler(kb, *keys)(event)

    assert mycli.key_bindings == expected_mode
    assert event.app.editing_mode == expected_editing_mode
    assert event.app.ttimeoutlen == expected_timeout


def test_tab_binding_uses_toolkit_default_to_start_completion() -> None:
    mycli = DummyMyCli(DummyKeysConfig(behaviors={'tab': ['toolkit_default']}))
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text='sel'))

    binding_handler(kb, Keys.ControlI)(event)

    assert event.app.current_buffer.start_completion_calls == [{'select_first': True, 'insert_common_part': False}]
    assert event.app.current_buffer.complete_next_calls == 0


def test_tab_binding_uses_toolkit_default_to_advance_existing_completion() -> None:
    mycli = DummyMyCli(DummyKeysConfig(behaviors={'tab': ['toolkit_default']}))
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text='sel', complete_state=object()))

    binding_handler(kb, Keys.ControlI)(event)

    assert event.app.current_buffer.complete_next_calls == 1


@pytest.mark.parametrize(
    ('behaviors', 'expected_start', 'expected_complete_next', 'expected_cancel'),
    (
        (['advance'], [], 1, 0),
        (['cancel'], [], 0, 1),
        (['advancing_summon'], [{'select_first': True, 'insert_common_part': False}], 0, 0),
        (['prefixing_summon'], [{'select_first': False, 'insert_common_part': True}], 0, 0),
        (['summon'], [{'select_first': False, 'insert_common_part': False}], 0, 0),
    ),
)
def test_tab_binding_supports_configured_behaviors(
    behaviors: list[str],
    expected_start: list[dict[str, bool]],
    expected_complete_next: int,
    expected_cancel: int,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(behaviors={'tab': behaviors}))
    kb = key_bindings.mycli_bindings(mycli)
    complete_state = object() if behaviors[0] in {'advance', 'cancel'} else None
    event = make_event(DummyBuffer(text='sel', complete_state=complete_state))

    binding_handler(kb, Keys.ControlI)(event)

    assert event.app.current_buffer.start_completion_calls == expected_start
    assert event.app.current_buffer.complete_next_calls == expected_complete_next
    assert event.app.current_buffer.cancel_completion_calls == expected_cancel


@pytest.mark.parametrize(
    ('editing_mode', 'expected_eager'),
    (
        (EditingMode.VI, True),
        (EditingMode.EMACS, False),
    ),
)
def test_escape_binding_cancels_completion_menu(
    monkeypatch,
    editing_mode: EditingMode,
    expected_eager: bool,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig())
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(complete_state=object()))
    event.app.editing_mode = editing_mode
    monkeypatch.setattr(key_bindings, 'get_app', lambda: event.app)
    patch_filter_app(monkeypatch, event.app)

    assert binding(kb, Keys.Escape).eager() is expected_eager
    assert binding_filter(kb, Keys.Escape)() is True

    inactive_event = make_event(DummyBuffer(complete_state=None))
    monkeypatch.setattr(key_bindings, 'get_app', lambda: inactive_event.app)
    assert binding_filter(kb, Keys.Escape)() is False

    monkeypatch.setattr(key_bindings, 'get_app', lambda: event.app)

    binding_handler(kb, Keys.Escape)(event)

    assert event.app.current_buffer.cancel_completion_calls == 1
    assert event.app.current_buffer.complete_state is None


def test_control_space_toolkit_default_starts_selection_for_non_empty_text() -> None:
    mycli = DummyMyCli(DummyKeysConfig(behaviors={'control_space': ['toolkit_default']}))
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text='abc'))

    binding_handler(kb, Keys.ControlAt)(event)

    assert event.app.current_buffer.start_selection_calls == [SelectionType.CHARACTERS]


def test_control_space_toolkit_default_is_noop_for_empty_text() -> None:
    mycli = DummyMyCli(DummyKeysConfig(behaviors={'control_space': ['toolkit_default']}))
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text=''))

    binding_handler(kb, Keys.ControlAt)(event)

    assert event.app.current_buffer.start_selection_calls == []
    assert event.app.current_buffer.start_completion_calls == []


@pytest.mark.parametrize(
    ('behaviors', 'expected_start', 'expected_complete_next', 'expected_cancel'),
    (
        (['advance'], [], 1, 0),
        (['cancel'], [], 0, 1),
        (['advancing_summon'], [{'select_first': True, 'insert_common_part': False}], 0, 0),
        (['prefixing_summon'], [{'select_first': False, 'insert_common_part': True}], 0, 0),
        (['summon'], [{'select_first': False, 'insert_common_part': False}], 0, 0),
    ),
)
def test_control_space_supports_completion_behaviors(
    behaviors: list[str],
    expected_start: list[dict[str, bool]],
    expected_complete_next: int,
    expected_cancel: int,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(behaviors={'control_space': behaviors}))
    kb = key_bindings.mycli_bindings(mycli)
    complete_state = object() if behaviors[0] in {'advance', 'cancel'} else None
    event = make_event(DummyBuffer(text='sel', complete_state=complete_state))

    binding_handler(kb, Keys.ControlAt)(event)

    assert event.app.current_buffer.start_completion_calls == expected_start
    assert event.app.current_buffer.complete_next_calls == expected_complete_next
    assert event.app.current_buffer.cancel_completion_calls == expected_cancel


@pytest.mark.parametrize(
    ('keys', 'handler_name'),
    (
        ((Keys.ControlX, 'p'), 'handle_prettify_binding'),
        ((Keys.ControlX, 'u'), 'handle_unprettify_binding'),
    ),
)
def test_prettify_bindings_transform_non_empty_buffer(
    monkeypatch,
    keys: tuple[str | Keys, ...],
    handler_name: str,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), key_bindings_mode='emacs')
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text='select 1'))
    event.app.editing_mode = EditingMode.EMACS
    patch_filter_app(monkeypatch, event.app)

    assert binding_filter(kb, *keys)() is True

    binding_handler(kb, *keys)(event)

    assert len(event.app.current_buffer.transform_calls) == 1
    start, end, handler = event.app.current_buffer.transform_calls[0]
    assert (start, end) == (0, len('select 1'))
    assert handler.func is getattr(key_bindings.key_binding_utils, handler_name)
    assert handler.args == (mycli,)


@pytest.mark.parametrize('keys', ((Keys.ControlX, 'p'), (Keys.ControlX, 'u')))
def test_prettify_bindings_skip_empty_buffer(monkeypatch, keys: tuple[str | Keys, ...]) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), key_bindings_mode='emacs')
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text=''))
    event.app.editing_mode = EditingMode.EMACS
    patch_filter_app(monkeypatch, event.app)

    binding_handler(kb, *keys)(event)

    assert event.app.current_buffer.transform_calls == []


@pytest.mark.parametrize(
    ('keys', 'expected_text'),
    (
        ((Keys.ControlO, 'd'), 'DATE'),
        ((Keys.ControlO, Keys.ControlD), "'DATE'"),
        ((Keys.ControlO, 't'), 'DATETIME'),
        ((Keys.ControlO, Keys.ControlT), "'DATETIME'"),
    ),
)
def test_date_and_datetime_bindings_insert_shortcuts(
    monkeypatch,
    keys: tuple[str | Keys, ...],
    expected_text: str,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), key_bindings_mode='emacs')
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event()
    event.app.editing_mode = EditingMode.EMACS
    patch_filter_app(monkeypatch, event.app)

    monkeypatch.setattr(
        key_bindings.key_binding_utils,
        'server_date',
        lambda _sqlexecute, quoted=False: "'DATE'" if quoted else 'DATE',
    )
    monkeypatch.setattr(
        key_bindings.key_binding_utils,
        'server_datetime',
        lambda _sqlexecute, quoted=False: "'DATETIME'" if quoted else 'DATETIME',
    )

    assert binding_filter(kb, *keys)() is True

    inactive_event = make_event()
    inactive_event.app.editing_mode = EditingMode.VI
    patch_filter_app(monkeypatch, inactive_event.app)
    assert binding_filter(kb, *keys)() is False

    patch_filter_app(monkeypatch, event.app)

    binding_handler(kb, *keys)(event)

    assert event.app.current_buffer.inserted_text == [expected_text]


def test_control_r_uses_reverse_isearch_mode_when_configured(monkeypatch) -> None:
    mycli = DummyMyCli(DummyKeysConfig(options={'control_r': 'reverse_isearch'}), key_bindings_mode='emacs')
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event()
    event.app.editing_mode = EditingMode.EMACS
    event.app.layout = SimpleNamespace(current_control=BufferControl(search_buffer_control=SearchBufferControl()))
    vi_mode_event = make_event()
    vi_mode_event.app.editing_mode = EditingMode.VI
    vi_mode_event.app.layout = SimpleNamespace(current_control=BufferControl(search_buffer_control=SearchBufferControl()))
    calls: list[dict[str, Any]] = []
    patch_filter_app(monkeypatch, event.app)

    monkeypatch.setattr(
        key_bindings,
        'search_history',
        lambda *args, **kwargs: calls.append({'args': args, 'kwargs': kwargs}),
    )

    assert binding_filter(kb, Keys.ControlR)() is True

    inactive_event = make_event()
    inactive_event.app.editing_mode = EditingMode.EMACS
    inactive_event.app.layout = SimpleNamespace(current_control=object())
    patch_filter_app(monkeypatch, inactive_event.app)
    assert binding_filter(kb, Keys.ControlR)() is False

    patch_filter_app(monkeypatch, vi_mode_event.app)
    assert binding_filter(kb, Keys.ControlR)() is True

    patch_filter_app(monkeypatch, event.app)

    binding_handler(kb, Keys.ControlR)(event)
    patch_filter_app(monkeypatch, vi_mode_event.app)
    binding_handler(kb, Keys.ControlR)(vi_mode_event)

    assert calls == [
        {'args': (event,), 'kwargs': {'incremental': True}},
        {'args': (vi_mode_event,), 'kwargs': {'incremental': True}},
    ]


def test_control_r_and_alt_r_use_fzf_search_options(monkeypatch) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), key_bindings_mode='emacs')
    kb = key_bindings.mycli_bindings(mycli)
    calls: list[dict[str, Any]] = []

    monkeypatch.setattr(
        key_bindings,
        'search_history',
        lambda *args, **kwargs: calls.append({'args': args, 'kwargs': kwargs}),
    )

    control_r_event = make_event()
    alt_r_event = make_event()
    control_r_event.app.editing_mode = EditingMode.EMACS
    alt_r_event.app.editing_mode = EditingMode.EMACS
    control_r_event.app.layout = SimpleNamespace(current_control=BufferControl(search_buffer_control=SearchBufferControl()))
    alt_r_event.app.layout = SimpleNamespace(current_control=BufferControl(search_buffer_control=SearchBufferControl()))
    patch_filter_app(monkeypatch, control_r_event.app)
    assert binding_filter(kb, Keys.ControlR)() is True

    inactive_control_r_event = make_event()
    inactive_control_r_event.app.editing_mode = EditingMode.EMACS
    inactive_control_r_event.app.layout = SimpleNamespace(current_control=object())
    patch_filter_app(monkeypatch, inactive_control_r_event.app)
    assert binding_filter(kb, Keys.ControlR)() is False

    vi_mode_control_r_event = make_event()
    vi_mode_control_r_event.app.editing_mode = EditingMode.VI
    vi_mode_control_r_event.app.layout = SimpleNamespace(current_control=BufferControl(search_buffer_control=SearchBufferControl()))
    patch_filter_app(monkeypatch, vi_mode_control_r_event.app)
    assert binding_filter(kb, Keys.ControlR)() is True

    patch_filter_app(monkeypatch, control_r_event.app)
    binding_handler(kb, Keys.ControlR)(control_r_event)
    patch_filter_app(monkeypatch, vi_mode_control_r_event.app)
    binding_handler(kb, Keys.ControlR)(vi_mode_control_r_event)
    patch_filter_app(monkeypatch, alt_r_event.app)
    assert binding_filter(kb, Keys.Escape, 'r')() is True

    vi_mode_event = make_event()
    vi_mode_event.app.editing_mode = EditingMode.VI
    vi_mode_event.app.layout = SimpleNamespace(current_control=BufferControl(search_buffer_control=SearchBufferControl()))
    patch_filter_app(monkeypatch, vi_mode_event.app)
    assert binding_filter(kb, Keys.Escape, 'r')() is False

    non_searchable_event = make_event()
    non_searchable_event.app.editing_mode = EditingMode.EMACS
    non_searchable_event.app.layout = SimpleNamespace(current_control=object())
    patch_filter_app(monkeypatch, non_searchable_event.app)
    assert binding_filter(kb, Keys.Escape, 'r')() is False

    patch_filter_app(monkeypatch, alt_r_event.app)
    binding_handler(kb, Keys.Escape, 'r')(alt_r_event)

    assert calls == [
        {
            'args': (control_r_event,),
            'kwargs': {
                'highlight_preview': True,
                'highlight_style': 'native',
            },
        },
        {
            'args': (vi_mode_control_r_event,),
            'kwargs': {
                'highlight_preview': True,
                'highlight_style': 'native',
            },
        },
        {
            'args': (alt_r_event,),
            'kwargs': {
                'highlight_preview': True,
                'highlight_style': 'native',
            },
        },
    ]


@pytest.mark.parametrize(
    ('mode', 'expected_exit_calls', 'expected_bells'),
    (
        ('exit', [{'exception': EOFError, 'style': 'class:exiting'}], 0),
        ('bell', [], 1),
    ),
)
def test_control_d_binding_exits_or_bells(
    monkeypatch,
    mode: str,
    expected_exit_calls: list[dict[str, Any]],
    expected_bells: int,
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(options={'control_d': mode}))
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event()
    monkeypatch.setattr(key_bindings, 'get_app', lambda: event.app)

    assert binding_filter(kb, Keys.ControlD)() is True

    inactive_event = make_event(DummyBuffer(text='select 1'))
    monkeypatch.setattr(key_bindings, 'get_app', lambda: inactive_event.app)
    assert binding_filter(kb, Keys.ControlD)() is False

    monkeypatch.setattr(key_bindings, 'get_app', lambda: event.app)

    binding_handler(kb, Keys.ControlD)(event)

    assert event.app.exit_calls == expected_exit_calls
    assert event.app.output.bell_calls == expected_bells


def test_enter_binding_closes_completion_menu(monkeypatch) -> None:
    mycli = DummyMyCli(DummyKeysConfig())
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event(DummyBuffer(text='sel', complete_state=SimpleNamespace(current_completion=object())))
    patch_filter_app(monkeypatch, event.app)

    assert binding_filter(kb, Keys.ControlM)() is True

    inactive_event = make_event(DummyBuffer(text='sel', complete_state=SimpleNamespace(current_completion=None)))
    patch_filter_app(monkeypatch, inactive_event.app)
    assert binding_filter(kb, Keys.ControlM)() is False

    patch_filter_app(monkeypatch, event.app)

    binding_handler(kb, Keys.ControlM)(event)

    assert event.current_buffer.complete_state is None
    assert event.app.current_buffer.complete_state is None


@pytest.mark.parametrize(
    ('multi_line', 'expected_validate_calls', 'expected_inserted_text'),
    (
        (True, 1, []),
        (False, 0, ['\n']),
    ),
)
def test_alt_enter_binding_validates_or_inserts_newline(
    multi_line: bool,
    expected_validate_calls: int,
    expected_inserted_text: list[str],
) -> None:
    mycli = DummyMyCli(DummyKeysConfig(), multi_line=multi_line)
    kb = key_bindings.mycli_bindings(mycli)
    event = make_event()

    binding_handler(kb, Keys.Escape, Keys.ControlM)(event)

    assert event.app.current_buffer.validate_calls == expected_validate_calls
    assert event.app.current_buffer.inserted_text == expected_inserted_text



from __future__ import annotations

from dataclasses import dataclass
from io import TextIOWrapper
import os
import sys
from typing import Callable

import click
import clickdc

import mycli as mycli_package
from mycli.cli_runner import run_from_cli_args
from mycli.client import MyCli
from mycli.constants import (
    DEFAULT_PROMPT,
    EMPTY_PASSWORD_FLAG_SENTINEL,
)
from mycli.packages.cli_utils import filtered_sys_argv


class IntOrStringClickParamType(click.ParamType):
    name = 'text'  # display as TEXT in helpdoc

    def convert(self, value, param, ctx):
        if isinstance(value, int):
            return value
        elif isinstance(value, str):
            return value
        elif value is None:
            return value
        else:
            self.fail('Not a valid password string', param, ctx)


INT_OR_STRING_CLICK_TYPE = IntOrStringClickParamType()


@dataclass(slots=True)
class CliArgs:
    database: str | None = clickdc.argument(
        type=str,
        default=None,
        nargs=1,
    )
    host: str | None = clickdc.option(
        '-h',
        '--hostname',
        'host',
        type=str,
        envvar='MYSQL_HOST',
        help='Host address of the database.',
    )
    port: int | None = clickdc.option(
        '-P',
        type=int,
        envvar='MYSQL_TCP_PORT',
        help='Port number to use for connection. Honors $MYSQL_TCP_PORT.',
    )
    user: str | None = clickdc.option(
        '-u',
        '--user',
        '--username',
        'user',
        type=str,
        envvar='MYSQL_USER',
        help='User name to connect to the database.',
    )
    socket: str | None = clickdc.option(
        '-S',
        type=str,
        envvar='MYSQL_UNIX_SOCKET',
        help='The socket file to use for connection.',
    )
    password: int | str | None = clickdc.option(
        '-p',
        '--pass',
        '--password',
        'password',
        type=INT_OR_STRING_CLICK_TYPE,
        is_flag=False,
        flag_value=EMPTY_PASSWORD_FLAG_SENTINEL,
        help='Prompt for (or pass in cleartext) the password to connect to the database.',
    )
    password_file: str | None = clickdc.option(
        type=click.Path(),
        help='File or FIFO path containing the password to connect to the db if not specified otherwise.',
    )
    ssh_user: str | None = clickdc.option(
        type=str,
        help='User name to connect to ssh server.',
    )
    ssh_host: str | None = clickdc.option(
        type=str,
        help='Host name to connect to ssh server.',
    )
    ssh_port: int = clickdc.option(
        type=int,
        default=22,
        help='Port to connect to ssh server.',
    )
    ssh_password: str | None = clickdc.option(
        type=str,
        help='Password to connect to ssh server.',
    )
    ssh_key_filename: str | None = clickdc.option(
        type=str,
        help='Private key filename (identify file) for the ssh connection.',
    )
    ssh_config_path: str = clickdc.option(
        type=str,
        help='Path to ssh configuration.',
        default=os.path.expanduser('~') + '/.ssh/config',
    )
    ssh_config_host: str | None = clickdc.option(
        type=str,
        help='Host to connect to ssh server reading from ssh configuration.',
    )
    list_ssh_config: bool = clickdc.option(
        is_flag=True,
        help='list ssh configurations in the ssh config (requires paramiko).',
    )
    ssh_warning_off: bool = clickdc.option(
        is_flag=True,
        help='Suppress the SSH deprecation notice.',
    )
    ssl_mode: str = clickdc.option(
        type=click.Choice(['auto', 'on', 'off']),
        help='Set desired SSL behavior. auto=preferred if TCP/IP, on=required, off=off.',
    )
    deprecated_ssl: bool | None = clickdc.option(
        '--ssl/--no-ssl',
        'deprecated_ssl',
        default=None,
        clickdc=None,
        help='Enable SSL for connection (automatically enabled with other flags).',
    )
    ssl_ca: str | None = clickdc.option(
        type=click.Path(exists=True),
        help='CA file in PEM format.',
    )
    ssl_capath: str | None = clickdc.option(
        type=click.Path(exists=True, file_okay=False, dir_okay=True),
        help='CA directory.',
    )
    ssl_cert: str | None = clickdc.option(
        type=click.Path(exists=True),
        help='X509 cert in PEM format.',
    )
    ssl_key: str | None = clickdc.option(
        type=click.Path(exists=True),
        help='X509 key in PEM format.',
    )
    ssl_cipher: str | None = clickdc.option(
        type=str,
        help='SSL cipher to use.',
    )
    tls_version: str | None = clickdc.option(
        type=click.Choice(['TLSv1', 'TLSv1.1', 'TLSv1.2', 'TLSv1.3'], case_sensitive=False),
        help='TLS protocol version for secure connection.',
    )
    ssl_verify_server_cert: bool = clickdc.option(
        is_flag=True,
        help=("""Verify server's "Common Name" in its cert against hostname used when connecting. This option is disabled by default."""),
    )
    verbose: int = clickdc.option(
        '-v',
        count=True,
        help='More verbose output and feedback.  Can be given multiple times.',
    )
    quiet: bool = clickdc.option(
        '-q',
        is_flag=True,
        help='Less verbose output and feedback.',
    )
    dbname: str | None = clickdc.option(
        '-D',
        '--database',
        'dbname',
        type=str,
        clickdc=None,
        help='Database or DSN to use for the connection.',
    )
    dsn: str = clickdc.option(
        '-d',
        type=str,
        default='',
        envvar='MYSQL_DSN',
        help='DSN alias configured in the ~/.myclirc file, or a full DSN.',
    )
    list_dsn: bool = clickdc.option(
        is_flag=True,
        help='Show list of DSN aliases configured in the [alias_dsn] section of ~/.myclirc.',
    )
    prompt: str | None = clickdc.option(
        '-R',
        type=str,
        help=f'Prompt format (Default: "{DEFAULT_PROMPT}").',
    )
    toolbar: str | None = clickdc.option(
        type=str,
        help='Toolbar format.',
    )
    logfile: TextIOWrapper | None = clickdc.option(
        '-l',
        type=click.File(mode='a', encoding='utf-8'),
        help='Log every query and its results to a file.',
    )
    checkpoint: TextIOWrapper | None = clickdc.option(
        type=click.File(mode='a', encoding='utf-8'),
        help='In batch or --execute mode, log successful queries to a file, and skipped with --resume.',
    )
    resume: bool = clickdc.option(
        '--resume',
        is_flag=True,
        help='In batch mode, resume after replaying statements in the --checkpoint file.',
    )
    defaults_group_suffix: str | None = clickdc.option(
        type=str,
        help='Read MySQL config groups with the specified suffix.',
    )
    defaults_file: str | None = clickdc.option(
        type=click.Path(),
        help='Only read MySQL options from the given file.',
    )
    myclirc: str = clickdc.option(
        type=click.Path(),
        default='~/.myclirc',
        help='Location of myclirc file.',
    )
    auto_vertical_output: bool = clickdc.option(
        is_flag=True,
        help='Automatically switch to vertical output mode if the result is wider than the terminal width.',
    )
    show_warnings: bool | None = clickdc.option(
        '--show-warnings/--no-show-warnings',
        is_flag=True,
        default=None,
        clickdc=None,
        help='Automatically show warnings after executing a SQL statement.',
    )
    table: bool = clickdc.option(
        '-t',
        is_flag=True,
        help='Shorthand for --format=table.',
    )
    csv: bool = clickdc.option(
        is_flag=True,
        help='Shorthand for --format=csv.',
    )
    warn: bool | None = clickdc.option(
        '--warn/--no-warn',
        default=None,
        clickdc=None,
        help='Warn before running a destructive query.',
    )
    local_infile: bool | None = clickdc.option(
        type=bool,
        is_flag=False,
        default=None,
        help='Enable/disable LOAD DATA LOCAL INFILE.',
    )
    login_path: str | None = clickdc.option(
        '-g',
        type=str,
        help='Read this path from the login file.',
    )
    execute: str | None = clickdc.option(
        '-e',
        type=str,
        help='Execute command and quit.',
    )
    init_command: str | None = clickdc.option(
        type=str,
        help='SQL statement to execute after connecting.',
    )
    unbuffered: bool | None = clickdc.option(
        is_flag=True,
        help='Instead of copying every row of data into a buffer, fetch rows as needed, to save memory.',
    )
    character_set: str | None = clickdc.option(
        '--charset',
        '--character-set',
        'character_set',
        type=str,
        help='Character set for MySQL session.',
    )
    batch: str | None = clickdc.option(
        type=str,
        help='SQL script to execute in batch mode.',
    )
    noninteractive: bool = clickdc.option(
        is_flag=True,
        help="Don't prompt during batch input.  Recommended.",
    )
    format: str | None = clickdc.option(
        type=click.Choice(['default', 'csv', 'tsv', 'table']),
        help='Format for batch or --execute output.',
    )
    throttle: float = clickdc.option(
        type=float,
        default=0.0,
        help='Pause in seconds between queries in batch mode.',
    )
    progress: bool = clickdc.option(
        is_flag=True,
        help='Show progress on the standard error with --batch.',
    )
    use_keyring: str | None = clickdc.option(
        type=click.Choice(['true', 'false', 'reset']),
        default=None,
        help='Store and retrieve passwords from the system keyring: true/false/reset.',
    )
    keepalive_ticks: int | None = clickdc.option(
        type=int,
        help='Send regular keepalive pings to the connection, roughly every <int> seconds.',
    )
    checkup: bool = clickdc.option(
        is_flag=True,
        help='Run a checkup on your configuration.',
    )


def get_password_from_file(password_file: str | None) -> str | None:
    if not password_file:
        return None
    try:
        with open(password_file) as fp:
            return fp.readline().removesuffix('\n')
    except FileNotFoundError:
        click.secho(f"Password file '{password_file}' not found", err=True, fg='red')
        sys.exit(1)
    except PermissionError:
        click.secho(f"Permission denied reading password file '{password_file}'", err=True, fg='red')
        sys.exit(1)
    except IsADirectoryError:
        click.secho(f"Path '{password_file}' is a directory, not a file", err=True, fg='red')
        sys.exit(1)
    except Exception as e:
        click.secho(f"Error reading password file '{password_file}': {str(e)}", err=True, fg='red')
        sys.exit(1)


def preprocess_cli_args(
    cli_args: CliArgs,
    is_valid_connection_scheme: Callable[[str], tuple[bool, str | None]],
) -> int:
    if cli_args.database is None and isinstance(cli_args.password, str) and '://' in cli_args.password:
        is_valid_scheme, scheme = is_valid_connection_scheme(cli_args.password)
        if not is_valid_scheme:
            click.secho(f'Error: Unknown connection scheme provided for DSN URI ({scheme}://)', err=True, fg='red')
            sys.exit(1)
        cli_args.database = cli_args.password
        cli_args.password = EMPTY_PASSWORD_FLAG_SENTINEL

    if cli_args.password is None and cli_args.password_file:
        password_from_file = get_password_from_file(cli_args.password_file)
        if password_from_file is not None:
            cli_args.password = password_from_file

    if cli_args.password is None and os.environ.get('MYSQL_PWD') is not None:
        cli_args.password = os.environ.get('MYSQL_PWD')

    if cli_args.resume and not cli_args.checkpoint:
        click.secho('Error: --resume requires a --checkpoint file.', err=True, fg='red')
        sys.exit(1)

    if cli_args.resume and not cli_args.batch:
        click.secho('Error: --resume requires a --batch file.', err=True, fg='red')
        sys.exit(1)

    if cli_args.verbose and cli_args.quiet:
        click.secho('Error: --verbose and --quiet are incompatible.', err=True, fg='red')
        sys.exit(1)
    elif cli_args.verbose:
        return int(cli_args.verbose)
    elif cli_args.quiet:
        return -1
    return 0


@click.command()
@clickdc.adddc('cli_args', CliArgs)
@click.version_option(mycli_package.__version__, '--version', '-V', help="Output mycli's version.")
def click_entrypoint(
    cli_args: CliArgs,
) -> None:
    """A MySQL terminal client with auto-completion and syntax highlighting.

    \b
    Examples:
      - mycli my_database
      - mycli -u my_user -h my_host.com my_database
      - mycli mysql://my_user@my_host.com:3306/my_database

    """

    run_from_cli_args(cli_args, client_factory=MyCli)


def main() -> int | None:
    try:
        result = click_entrypoint.main(
            filtered_sys_argv(),
            standalone_mode=False,  # disable builtin exception handling
            prog_name='mycli',
        )
    except click.Abort:
        print('Aborted!', file=sys.stderr)
        sys.exit(1)
    except BrokenPipeError:
        sys.exit(1)
    except click.ClickException as e:
        e.show()
        if hasattr(e, 'exit_code'):
            sys.exit(e.exit_code)
        else:
            sys.exit(2)
    if result is None:
        return 0
    elif isinstance(result, int):
        return result
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main())



from dataclasses import dataclass
import functools
import re
from typing import Any, Callable, Literal

import sqlparse
from sqlparse.sql import Comparison, Identifier, Token, Where

from mycli.packages.special.main import COMMANDS as SPECIAL_COMMANDS
from mycli.packages.special.main import parse_special_command
from mycli.packages.sql_utils import extract_tables, find_prev_keyword, last_word

sqlparse.engine.grouping.MAX_GROUPING_DEPTH = None  # type: ignore[assignment]
sqlparse.engine.grouping.MAX_GROUPING_TOKENS = None  # type: ignore[assignment]

_ENUM_VALUE_RE = re.compile(
    r"(?P<lhs>(?:`[^`]+`|[\w$]+)(?:\.(?:`[^`]+`|[\w$]+))?)\s*=\s*$",
    re.IGNORECASE,
)

# missing because not binary
#   BETWEEN
#   CASE
# missing because parens are used
#   IN(), and others
# unary operands might need to have another set
#   not, !, ~
# arrow operators only take a literal on the right
#   and so might need different treatment
# := might also need a different context
# sqlparse would call these identifiers, so they are excluded
#   xor
# these are hitting the recursion guard, and so not completing after
# so we might as well leave them out:
#   is, 'is not', mod
# sqlparse might also parse "not null" together
# should also verify how sqlparse parses every space-containing case
BINARY_OPERANDS = {
    '&', '>', '>>', '>=', '<', '<>', '!=', '<<', '<=', '<=>', '%',
    '*', '+', '-', '->', '->>', '/', ':=', '=', '^', 'and', '&&', 'div',
    'like', 'not like', 'not regexp', 'or', '||', 'regexp', 'rlike',
    'sounds like', '|',
}  # fmt: skip

Suggestion = dict[str, Any]
Predicate = Callable[['SuggestContext'], bool]
Emitter = Callable[['SuggestContext'], list[Suggestion]]


@dataclass(frozen=True)
class SuggestContext:
    token: str | Token | None
    token_value: str | None
    text_before_cursor: str
    word_before_cursor: str | None
    full_text: str
    identifier: Identifier
    parsed_cb: Callable[[], sqlparse.sql.Statement]
    tokens_wo_space_cb: Callable[[], list[Token]]


@dataclass(frozen=True)
class SuggestRule:
    name: str
    predicate: Predicate
    emit: Emitter


def _keyword_suggestions() -> list[Suggestion]:
    return [{'type': 'keyword'}]


def _keyword_and_special_suggestions() -> list[Suggestion]:
    return [{'type': 'keyword'}, {'type': 'special'}]


@functools.lru_cache(maxsize=128)
def _parse_suggestion_statement(text_before_cursor: str) -> sqlparse.sql.Statement:
    try:
        return sqlparse.parse(text_before_cursor)[0]
    except (AttributeError, IndexError, ValueError, sqlparse.exceptions.SQLParseError):
        return sqlparse.sql.Statement()


@functools.lru_cache(maxsize=128)
def _tokens_wo_space(text_before_cursor: str) -> list[Token]:
    parsed = _parse_suggestion_statement(text_before_cursor)
    return [x for x in parsed.tokens if x.ttype != sqlparse.tokens.Token.Text.Whitespace]


def _normalize_token_value(token: str | Token | None) -> str | None:
    if isinstance(token, str):
        return token.lower()
    if isinstance(token, Comparison):
        # If 'token' is a Comparison type such as
        # 'select * FROM abc a JOIN def d ON a.id = d.'. Then calling
        # token.value on the comparison type will only return the lhs of the
        # comparison. In this case a.id. So we need to do token.tokens to get
        # both sides of the comparison and pick the last token out of that
        # list.
        return token.tokens[-1].value.lower()
    if token is None:
        return None
    return token.value.lower()


def _build_suggest_context(
    token: str | Token | None,
    text_before_cursor: str,
    word_before_cursor: str | None,
    full_text: str,
    identifier: Identifier,
) -> SuggestContext:
    return SuggestContext(
        token=token,
        token_value=_normalize_token_value(token),
        text_before_cursor=text_before_cursor,
        word_before_cursor=word_before_cursor,
        full_text=full_text,
        identifier=identifier,
        parsed_cb=functools.partial(_parse_suggestion_statement, text_before_cursor),
        tokens_wo_space_cb=functools.partial(_tokens_wo_space, text_before_cursor),
    )


def _is_single_or_double_quoted(ctx: SuggestContext) -> bool:
    return is_inside_quotes(ctx.text_before_cursor, -1) in ['single', 'double']


def _parent_name(ctx: SuggestContext) -> str | list[Any]:
    return (ctx.identifier and ctx.identifier.get_parent_name()) or []


def _tables(ctx: SuggestContext) -> list[tuple[str | None, str, str]]:
    return extract_tables(ctx.full_text)


def _aliases(tables: list[tuple[str | None, str, str]]) -> list[str]:
    return [alias or table for (schema, table, alias) in tables]


def _emit_none_token(_ctx: SuggestContext) -> list[Suggestion]:
    return _keyword_suggestions()


def _emit_blank_token(_ctx: SuggestContext) -> list[Suggestion]:
    return _keyword_and_special_suggestions()


def _emit_star(_ctx: SuggestContext) -> list[Suggestion]:
    return _keyword_suggestions()


def _emit_lparen(ctx: SuggestContext) -> list[Suggestion]:
    if ctx.parsed_cb().tokens and isinstance(ctx.parsed_cb().tokens[-1], Where):
        # Four possibilities:
        #  1 - Parenthesized clause like "WHERE foo AND ("
        #        Suggest columns/functions
        #  2 - Function call like "WHERE foo("
        #        Suggest columns/functions
        #  3 - Subquery expression like "WHERE EXISTS ("
        #        Suggest keywords, in order to do a subquery
        #  4 - Subquery OR array comparison like "WHERE foo = ANY("
        #        Suggest columns/functions AND keywords. (If we wanted to be
        #        really fancy, we could suggest only array-typed columns)

        # override a few properties in the SuggestContext
        column_suggestions = _emit_select_like(
            SuggestContext(
                token='where',
                token_value='where',
                text_before_cursor=ctx.text_before_cursor,
                word_before_cursor=None,
                full_text=ctx.full_text,
                identifier=ctx.identifier,
                parsed_cb=ctx.parsed_cb,
                tokens_wo_space_cb=ctx.tokens_wo_space_cb,
            )
        )

        # Check for a subquery expression (cases 3 & 4)
        where = ctx.parsed_cb().tokens[-1]
        _idx, prev_tok = where.token_prev(len(where.tokens) - 1)

        if isinstance(prev_tok, Comparison):
            # e.g. "SELECT foo FROM bar WHERE foo = ANY("
            prev_tok = prev_tok.tokens[-1]

        prev_tok = prev_tok.value.lower()
        if prev_tok == 'exists':
            return _keyword_suggestions()
        return column_suggestions

    # Get the token before the parens
    _idx, prev_tok = ctx.parsed_cb().token_prev(len(ctx.parsed_cb().tokens) - 1)
    if prev_tok and prev_tok.value and prev_tok.value.lower() == 'using':
        # tbl1 INNER JOIN tbl2 USING (col1, col2)
        # suggest columns that are present in more than one table
        return [{'type': 'column', 'tables': _tables(ctx), 'drop_unique': True}]
    if ctx.parsed_cb().tokens and ctx.parsed_cb().token_first() and ctx.parsed_cb().token_first().value.lower() == 'select':
        # If the lparen is preceeded by a space chances are we're about to
        # do a sub-select.
        if last_word(ctx.text_before_cursor, 'all_punctuations').startswith('('):
            return _keyword_suggestions()
    elif ctx.parsed_cb().tokens and ctx.parsed_cb().token_first() and ctx.parsed_cb().token_first().value.lower() == 'show':
        return [{'type': 'show'}]

    # We're probably in a function argument list
    return [{'type': 'column', 'tables': _tables(ctx)}]


def _emit_procedure(_ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'procedure', 'schema': []}]


def _emit_character_set(_ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'character_set'}]


def _emit_column_for_tables(ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'column', 'tables': _tables(ctx)}]


def _emit_nothing(_ctx: SuggestContext) -> list[Suggestion]:
    return []


def _emit_show(_ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'show'}]


def _emit_to(ctx: SuggestContext) -> list[Suggestion]:
    if ctx.parsed_cb().tokens and ctx.parsed_cb().token_first() and ctx.parsed_cb().token_first().value.lower() == 'change':
        return [{'type': 'change'}]
    return [{'type': 'user'}]


def _emit_user(_ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'user'}]


def _emit_collation(_ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'collation'}]


def _emit_select_like(ctx: SuggestContext) -> list[Suggestion]:
    parent = _parent_name(ctx)
    tables = _tables(ctx)
    if parent:
        tables = [t for t in tables if identifies(parent, *t)]
        return [
            {'type': 'column', 'tables': tables},
            {'type': 'table', 'schema': parent},
            {'type': 'view', 'schema': parent},
            {'type': 'function', 'schema': parent},
        ]
    if is_inside_quotes(ctx.text_before_cursor, -1) == 'backtick':
        # todo: this should be revised, since we complete too exuberantly within
        # backticks, including keywords
        aliases = _aliases(tables)
        return [
            {'type': 'column', 'tables': tables},
            {'type': 'function', 'schema': []},
            {'type': 'alias', 'aliases': aliases},
            {'type': 'keyword'},
        ]

    aliases = _aliases(tables)
    return [
        {'type': 'column', 'tables': tables},
        {'type': 'function', 'schema': []},
        {'type': 'introducer'},
        {'type': 'alias', 'aliases': aliases},
    ]


def _emit_relation_like(ctx: SuggestContext) -> list[Suggestion]:
    schema = _parent_name(ctx)
    is_join = bool(ctx.token_value and ctx.token_value.endswith('join') and isinstance(ctx.token, Token) and ctx.token.is_keyword)

    # Suggest tables from either the currently-selected schema or the
    # public schema if no schema has been specified
    table_suggestion: Suggestion = {'type': 'table', 'schema': schema}
    if is_join:
        table_suggestion['join'] = True
    suggest: list[Suggestion] = [table_suggestion]

    if not schema:
        # Suggest schemas
        suggest.append({'type': 'database'})

    # Only tables can be TRUNCATED, otherwise suggest views
    if ctx.token_value != 'truncate':
        suggest.append({'type': 'view', 'schema': schema})

    return suggest


def _emit_relation_name(ctx: SuggestContext) -> list[Suggestion]:
    rel_type = ctx.token_value
    assert rel_type is not None
    schema = _parent_name(ctx)
    if schema:
        return [{'type': rel_type, 'schema': schema}]
    return [{'type': 'schema'}, {'type': rel_type, 'schema': []}]


def _emit_on(ctx: SuggestContext) -> list[Suggestion]:
    tables = _tables(ctx)  # [(schema, table, alias), ...]
    parent = _parent_name(ctx)
    if parent:
        # "ON parent.<suggestion>"
        # parent can be either a schema name or table alias
        # todo recognize and separate schema and table suggestions
        # todo remove function suggestions here
        tables = [t for t in tables if identifies(parent, *t)]
        return [
            {'type': 'column', 'tables': tables},
            {'type': 'table', 'schema': parent},
            {'type': 'view', 'schema': parent},
            {'type': 'function', 'schema': parent},
        ]

    # ON <suggestion>
    # Use table alias if there is one, otherwise the table name
    aliases = _aliases(tables)
    suggest: list[Suggestion] = [{'type': 'fk_join', 'tables': tables}, {'type': 'alias', 'aliases': aliases}]

    # The lists of 'aliases' could be empty if we're trying to complete
    # a GRANT query. eg: GRANT SELECT, INSERT ON <tab>
    # In that case we just suggest all schemata and all tables.
    if not aliases:
        suggest.append({'type': 'database'})
        suggest.append({'type': 'table', 'schema': parent})
    return suggest


def _emit_database(_ctx: SuggestContext) -> list[Suggestion]:
    return [{'type': 'database'}]


def _emit_where_token(ctx: SuggestContext) -> list[Suggestion]:
    assert isinstance(ctx.token, Where)
    # sqlparse groups all tokens from the where clause into a single token
    # list. This means that token.value may be something like
    # 'where foo > 5 and '. We need to look "inside" token.tokens to handle
    # suggestions in complicated where clauses correctly.
    #
    # This logic also needs to look even deeper in to the WHERE clause.
    # We recapitulate some transcoding suggestions here, but cannot
    # recapitulate the entire logic of this function.
    where_tokens = [x for x in ctx.token.tokens if x.ttype != sqlparse.tokens.Token.Text.Whitespace]
    if transcoding_suggestion := _charset_suggestion(where_tokens):
        return transcoding_suggestion

    original_text = ctx.text_before_cursor
    prev_keyword, rewound_text = find_prev_keyword(ctx.text_before_cursor)
    enum_suggestion = _enum_value_suggestion(original_text, ctx.full_text)
    fallback = suggest_based_on_last_token(prev_keyword, rewound_text, None, ctx.full_text, ctx.identifier)
    if enum_suggestion and _is_where_or_having(prev_keyword):
        return [enum_suggestion] + fallback
    return fallback


def _emit_binary_or_comma(ctx: SuggestContext) -> list[Suggestion]:
    original_text = ctx.text_before_cursor
    prev_keyword, rewound_text = find_prev_keyword(ctx.text_before_cursor)
    enum_suggestion = _enum_value_suggestion(original_text, ctx.full_text)

    # guard against non-progressing parser rewinds, which can otherwise
    # recurse forever on some operator shapes.
    if prev_keyword and rewound_text.rstrip() != original_text.rstrip():
        fallback = suggest_based_on_last_token(prev_keyword, rewound_text, None, ctx.full_text, ctx.identifier)
    else:
        # perhaps this fallback should include columns
        fallback = _keyword_suggestions()

    if enum_suggestion and _is_where_or_having(prev_keyword):
        return [enum_suggestion] + fallback
    return fallback


def _word_starts_with_digit_or_dot(ctx: SuggestContext) -> bool:
    return bool(ctx.word_before_cursor and re.match(r'^[\d\.]', ctx.word_before_cursor[0]))


def _word_starts_with_quote(ctx: SuggestContext) -> bool:
    return bool(ctx.word_before_cursor and ctx.word_before_cursor[0] in ('"', "'"))


def _word_inside_single_or_double_quotes(ctx: SuggestContext) -> bool:
    return bool(ctx.word_before_cursor and _is_single_or_double_quoted(ctx))


def _token_is_none(ctx: SuggestContext) -> bool:
    return ctx.token is None


def _token_is_blank(ctx: SuggestContext) -> bool:
    return not ctx.token


def _token_value_is(ctx: SuggestContext, *values: str) -> bool:
    return bool(ctx.token_value and ctx.token_value in values)


def _token_is_lparen(ctx: SuggestContext) -> bool:
    return bool(ctx.token_value and ctx.token_value.endswith('('))


def _token_is_relation_keyword(ctx: SuggestContext) -> bool:
    return bool(
        (ctx.token_value and ctx.token_value.endswith('join') and isinstance(ctx.token, Token) and ctx.token.is_keyword)
        or (ctx.token_value in ('copy', 'from', 'update', 'into', 'describe', 'truncate', 'desc', 'explain'))
        or (ctx.token_value == 'like' and re.match(r'^\s*create\s+table\s', ctx.full_text, re.IGNORECASE))
    )


def _token_is_binary_or_comma(ctx: SuggestContext) -> bool:
    return bool(ctx.token_value and (ctx.token_value.endswith(',') or ctx.token_value in BINARY_OPERANDS))


SUGGEST_BASED_ON_LAST_TOKEN_RULES = [
    SuggestRule(
        'guard_number_or_dot',
        _word_starts_with_digit_or_dot,
        _emit_nothing,
    ),
    SuggestRule(
        'guard_quote_prefix',
        _word_starts_with_quote,
        _emit_nothing,
    ),
    SuggestRule(
        'guard_inside_single_or_double',
        _word_inside_single_or_double_quotes,
        _emit_nothing,
    ),
    SuggestRule(
        'where_token',
        lambda ctx: isinstance(ctx.token, Where),
        _emit_where_token,
    ),
    SuggestRule(
        'none_token',
        _token_is_none,
        _emit_none_token,
    ),
    SuggestRule(
        'blank_token',
        _token_is_blank,
        _emit_blank_token,
    ),
    SuggestRule(
        'star_token',
        lambda ctx: _token_value_is(ctx, '*'),
        _emit_star,
    ),
    SuggestRule(
        'lparen_token',
        _token_is_lparen,
        _emit_lparen,
    ),
    SuggestRule(
        'call',
        lambda ctx: _token_value_is(ctx, 'call'),
        _emit_procedure,
    ),
    SuggestRule(
        'character_set_after_character',
        lambda ctx: (
            _token_value_is(ctx, 'set') and len(ctx.tokens_wo_space_cb()) >= 3 and ctx.tokens_wo_space_cb()[-3].value.lower() == 'character'
        ),
        _emit_character_set,
    ),
    SuggestRule(
        'character_set_after_character_short',
        lambda ctx: (
            _token_value_is(ctx, 'set') and len(ctx.tokens_wo_space_cb()) >= 2 and ctx.tokens_wo_space_cb()[-2].value.lower() == 'character'
        ),
        _emit_character_set,
    ),
    SuggestRule(
        'set_order_by_distinct',
        lambda ctx: _token_value_is(ctx, 'set', 'order by', 'distinct'),
        _emit_column_for_tables,
    ),
    SuggestRule(
        'as',
        lambda ctx: _token_value_is(ctx, 'as'),
        _emit_nothing,
    ),
    SuggestRule(
        'show',
        lambda ctx: _token_value_is(ctx, 'show'),
        _emit_show,
    ),
    SuggestRule(
        'to',
        lambda ctx: _token_value_is(ctx, 'to'),
        _emit_to,
    ),
    SuggestRule(
        'user_or_for',
        lambda ctx: _token_value_is(ctx, 'user', 'for'),
        _emit_user,
    ),
    SuggestRule(
        'collate',
        lambda ctx: _token_value_is(ctx, 'collate'),
        _emit_collation,
    ),
    SuggestRule(
        'using_after_convert_long',
        lambda ctx: (
            _token_value_is(ctx, 'using') and len(ctx.tokens_wo_space_cb()) >= 5 and ctx.tokens_wo_space_cb()[-5].value.lower() == 'convert'
        ),
        _emit_character_set,
    ),
    SuggestRule(
        'using_after_convert_short',
        lambda ctx: (
            _token_value_is(ctx, 'using') and len(ctx.tokens_wo_space_cb()) >= 4 and ctx.tokens_wo_space_cb()[-4].value.lower() == 'convert'
        ),
        _emit_character_set,
    ),
    SuggestRule(
        'select_where_having',
        lambda ctx: _token_value_is(ctx, 'select', 'where', 'having'),
        _emit_select_like,
    ),
    SuggestRule(
        'relation_keyword',
        _token_is_relation_keyword,
        _emit_relation_like,
    ),
    SuggestRule(
        'relation_name',
        lambda ctx: _token_value_is(ctx, 'table', 'view', 'function'),
        _emit_relation_name,
    ),
    SuggestRule(
        'on',
        lambda ctx: _token_value_is(ctx, 'on'),
        _emit_on,
    ),
    SuggestRule(
        'database_template',
        lambda ctx: _token_value_is(ctx, 'database', 'template'),
        _emit_database,
    ),
    SuggestRule(
        'inside_single_or_double',
        _is_single_or_double_quoted,
        _emit_nothing,
    ),
    SuggestRule(
        'binary_or_comma',
        _token_is_binary_or_comma,
        _emit_binary_or_comma,
    ),
]


def _enum_value_suggestion(text_before_cursor: str, full_text: str) -> dict[str, Any] | None:
    match = _ENUM_VALUE_RE.search(text_before_cursor)
    if not match:
        return None
    if is_inside_quotes(text_before_cursor, match.start("lhs")):
        return None

    lhs = match.group("lhs")
    if "." in lhs:
        parent, column = lhs.split(".", 1)
    else:
        parent, column = None, lhs

    return {
        "type": "enum_value",
        "tables": extract_tables(full_text),
        "column": column,
        "parent": parent,
    }


def _charset_suggestion(tokens: list[Token]) -> list[dict[str, str]] | None:
    token_values = [token.value.lower() for token in tokens if token.value]

    if len(token_values) >= 2 and token_values[-1] == 'set' and token_values[-2] == 'character':
        return [{'type': 'character_set'}]
    if len(token_values) >= 3 and token_values[-2] == 'set' and token_values[-3] == 'character':
        return [{'type': 'character_set'}]
    if len(token_values) >= 5 and token_values[-1] == 'using' and token_values[-4] == 'convert':
        return [{'type': 'character_set'}]
    if len(token_values) >= 6 and token_values[-2] == 'using' and token_values[-5] == 'convert':
        return [{'type': 'character_set'}]
    if len(token_values) >= 1 and token_values[-1] == 'collate':
        return [{'type': 'collation'}]

    return None


def _is_where_or_having(token: Token | None) -> bool:
    return bool(token and token.value and token.value.lower() in ("where", "having"))


def _find_doubled_backticks(text: str) -> list[int]:
    length = len(text)
    doubled_backtick_positions: list[int] = []
    backtick = '`'
    two_backticks = backtick + backtick

    if two_backticks not in text:
        return doubled_backtick_positions

    for index in range(0, length):
        ch = text[index]
        if ch != backtick:
            index += 1
            continue
        if index + 1 < length and text[index + 1] == backtick:
            doubled_backtick_positions.append(index)
            doubled_backtick_positions.append(index + 1)
            index += 2
            continue
        index += 1

    return doubled_backtick_positions


@functools.lru_cache(maxsize=128)
def is_inside_quotes(text: str, pos: int) -> Literal[False, 'single', 'double', 'backtick']:
    in_single = False
    in_double = False
    in_backticks = False
    escaped = False
    doubled_backtick_positions = []
    single_quote = "'"
    double_quote = '"'
    backtick = '`'
    backslash = '\\'

    # scanning the string twice seems to be needed to handle doubled backticks
    doubled_backtick_positions = _find_doubled_backticks(text)

    length = len(text)
    if pos < 0:
        pos = length + pos
        pos = max(pos, 0)
    pos = min(length, pos)

    # optimization
    up_to_pos = text[:pos]
    if backtick not in up_to_pos and single_quote not in up_to_pos and double_quote not in up_to_pos:
        return False

    for index in range(0, pos):
        ch = text[index]
        if index in doubled_backtick_positions:
            index += 1
            continue
        if escaped and (in_double or in_single):
            escaped = False
            index += 1
            continue
        if ch == backslash and (in_double or in_single):
            escaped = True
            index += 1
            continue
        if ch == backtick and not in_double and not in_single:
            in_backticks = not in_backticks
        elif ch == single_quote and not in_double and not in_backticks:
            in_single = not in_single
        elif ch == double_quote and not in_single and not in_backticks:
            in_double = not in_double
        index += 1

    if in_single:
        return 'single'
    elif in_double:
        return 'double'
    elif in_backticks:
        return 'backtick'
    else:
        return False


def suggest_type(full_text: str, text_before_cursor: str) -> list[dict[str, Any]]:
    """Takes the full_text that is typed so far and also the text before the
    cursor to suggest completion type and scope.

    Returns a tuple with a type of entity ('table', 'column' etc) and a scope.
    A scope for a column category will be a list of tables.
    """

    word_before_cursor = last_word(text_before_cursor, include="many_punctuations")

    identifier: Identifier | None = None

    # here should be removed once sqlparse has been fixed
    try:
        # If we've partially typed a word then word_before_cursor won't be an empty
        # string. In that case we want to remove the partially typed string before
        # sending it to the sqlparser. Otherwise the last token will always be the
        # partially typed string which renders the smart completion useless because
        # it will always return the list of keywords as completion.
        if word_before_cursor:
            if word_before_cursor.endswith("(") or word_before_cursor.startswith("\\"):
                parsed = sqlparse.parse(text_before_cursor)
            else:
                parsed = sqlparse.parse(text_before_cursor[: -len(word_before_cursor)])

                # word_before_cursor may include a schema qualification, like
                # "schema_name.partial_name" or "schema_name.", so parse it
                # separately
                p = sqlparse.parse(word_before_cursor)[0]

                if p.tokens and isinstance(p.tokens[0], Identifier):
                    identifier = p.tokens[0]
        else:
            parsed = sqlparse.parse(text_before_cursor)
    except (TypeError, AttributeError):
        return [{"type": "keyword"}]

    if len(parsed) > 1:
        # Multiple statements being edited -- isolate the current one by
        # cumulatively summing statement lengths to find the one that bounds the
        # current position
        current_pos = len(text_before_cursor)
        stmt_start, stmt_end = 0, 0

        for statement in parsed:
            stmt_len = len(str(statement))
            stmt_start, stmt_end = stmt_end, stmt_end + stmt_len

            if stmt_end >= current_pos:
                text_before_cursor = full_text[stmt_start:current_pos]
                full_text = full_text[stmt_start:]
                break

    elif parsed:
        # A single statement
        statement = parsed[0]
    else:
        # The empty string
        statement = None

    # Check for special commands and handle those separately
    if statement:
        # Be careful here because trivial whitespace is parsed as a statement,
        # but the statement won't have a first token
        tok1 = statement.token_first()
        # lenient because \. will parse as two tokens
        if tok1 and tok1.value.startswith('\\'):
            return suggest_special(text_before_cursor)
        elif tok1:
            if tok1.value.lower() in SPECIAL_COMMANDS:
                return suggest_special(text_before_cursor)

    last_token = statement and statement.token_prev(len(statement.tokens))[1] or ""

    # todo: unsure about empty string as identifier
    return suggest_based_on_last_token(last_token, text_before_cursor, word_before_cursor, full_text, identifier or Identifier(''))


def suggest_special(text: str) -> list[dict[str, Any]]:
    text = text.lstrip()
    cmd, _separator, _arg = parse_special_command(text)

    if cmd == text:
        # Trying to complete the special command itself
        return [{"type": "special"}]

    if cmd in ("\\u", "\\r"):
        return [{"type": "database"}]

    if cmd.lower() in ('use', 'connect'):
        return [{'type': 'database'}]

    if cmd in (r'\T', r'\Tr'):
        return [{"type": "table_format"}]

    if cmd.lower() in ('tableformat', 'redirectformat'):
        return [{"type": "table_format"}]

    if cmd in ["\\f", "\\fs", "\\fd"]:
        return [{"type": "favoritequery"}]

    if cmd in ["\\dt", "\\dt+"]:
        return [
            {"type": "table", "schema": []},
            {"type": "view", "schema": []},
            {"type": "schema"},
        ]
    elif cmd.lower() in [
        r'\.',
        'source',
        r'\o',
        r'\once',
        r'tee',
    ]:
        return [{"type": "file_name"}]
    # todo: why is \edit case-sensitive?
    elif cmd in [
        r'\e',
        r'\edit',
    ]:
        return [{"type": "file_name"}]
    if cmd in ["\\llm", "\\ai"]:
        return [{"type": "llm"}]

    return [{"type": "keyword"}, {"type": "special"}]


def suggest_based_on_last_token(
    token: str | Token | None,
    text_before_cursor: str,
    word_before_cursor: str | None,
    full_text: str,
    identifier: Identifier,
) -> list[dict[str, Any]]:
    ctx = _build_suggest_context(token, text_before_cursor, word_before_cursor, full_text, identifier)
    for rule in SUGGEST_BASED_ON_LAST_TOKEN_RULES:
        if rule.predicate(ctx):
            return rule.emit(ctx)

    return _keyword_suggestions()


def identifies(
    identifier: Any,
    schema: str | None,
    table: str,
    alias: str,
) -> bool:
    if identifier == alias:
        return True
    if identifier == table:
        return True
    if schema and identifier == (schema + "." + table):
        return True
    return False



import logging
import os
import platform

from pymysql import ProgrammingError
from pymysql.cursors import Cursor

from mycli import __version__
from mycli.packages.special import iocommands
from mycli.packages.special.main import ArgType, SpecialCommandAlias, special_command
from mycli.packages.special.utils import (
    format_uptime,
    get_local_timezone,
    get_server_timezone,
    get_ssl_cipher,
    get_ssl_version,
)
from mycli.packages.sqlresult import SQLResult

logger = logging.getLogger(__name__)


@special_command(
    "\\dt",
    "\\dt[+] [table]",
    "List or describe tables.",
    arg_type=ArgType.PARSED_QUERY,
    case_sensitive=True,
)
def list_tables(
    cur: Cursor,
    arg: str | None = None,
    _arg_type: ArgType = ArgType.PARSED_QUERY,
    command_verbosity: bool = False,
) -> list[SQLResult]:
    if arg:
        query = f'SHOW FIELDS FROM {arg}'
    else:
        query = "SHOW TABLES"
    logger.debug(query)
    cur.execute(query)
    if cur.description:
        header = [x[0] for x in cur.description]
    else:
        return [SQLResult()]

    # Fetch results before potentially executing another query
    results = list(cur.fetchall()) if command_verbosity and arg else cur

    postamble = ''
    if command_verbosity and arg:
        query = f'SHOW CREATE TABLE {arg}'
        logger.debug(query)
        cur.execute(query)
        if one := cur.fetchone():
            postamble = one[1]

    # todo missing a status line because sqlexecute.get_result was not used
    return [SQLResult(header=header, rows=results, postamble=postamble)]


@special_command(
    "\\l",
    "\\l",
    "List databases.",
    arg_type=ArgType.RAW_QUERY,
    case_sensitive=True,
)
def list_databases(cur: Cursor, **_) -> list[SQLResult]:
    query = "SHOW DATABASES"
    logger.debug(query)
    cur.execute(query)
    if cur.description:
        header = [x[0] for x in cur.description]
        # todo missing a status line because sqlexecute.get_result was not used
        return [SQLResult(header=header, rows=cur)]
    else:
        return [SQLResult()]


@special_command(
    "status",
    "status",
    "Get status information from the server.",
    arg_type=ArgType.RAW_QUERY,
    case_sensitive=True,
    aliases=[SpecialCommandAlias("\\s", case_sensitive=True)],
)
def status(cur: Cursor, **_) -> list[SQLResult]:
    query = "SHOW GLOBAL STATUS;"
    logger.debug(query)
    try:
        cur.execute(query)
    except ProgrammingError:
        # Fallback in case query fails, as it does with Mysql 4
        query = "SHOW STATUS;"
        logger.debug(query)
        cur.execute(query)
    status = dict(cur.fetchall())

    query = "SHOW GLOBAL VARIABLES;"
    logger.debug(query)
    cur.execute(query)
    global_variables = dict(cur.fetchall())

    query = "SHOW SESSION VARIABLES;"
    logger.debug(query)
    cur.execute(query)
    session_variables = dict(cur.fetchall())

    # decode in case keys are bytes, as with Mysql 4
    if global_variables and isinstance(list(global_variables)[0], bytes):
        global_variables = {k.decode("utf-8"): v.decode("utf-8") for k, v in global_variables.items()}
    if session_variables and isinstance(list(session_variables)[0], bytes):
        session_variables = {k.decode("utf-8"): v.decode("utf-8") for k, v in session_variables.items()}
    if status and isinstance(list(status)[0], bytes):
        status = {k.decode("utf-8"): v.decode("utf-8") for k, v in status.items()}

    # Create output buffers.
    preamble = []
    header = ['Setting', 'Value']
    output = []
    footer = []

    preamble.append("--------------")

    # Output the mycli client information.
    implementation = platform.python_implementation()
    version = platform.python_version()
    client_info = []
    client_info.append(f'mycli {__version__}')
    client_info.append(f'running on {implementation} {version}')
    preamble.append(" ".join(client_info) + "\n")

    # Build the output that will be displayed as a table.
    output.append(("Connection id:", cur.connection.thread_id()))

    query = "SELECT DATABASE(), USER();"
    logger.debug(query)
    cur.execute(query)
    if one := cur.fetchone():
        db, user = one
    else:
        db = ""
        user = ""
    output.append(("Current database:", db))
    output.append(("Current user:", user))

    if iocommands.is_pager_enabled():
        if "PAGER" in os.environ:
            pager = os.environ["PAGER"]
        else:
            pager = "System default"
    else:
        pager = "stdout"
    output.append(("Current pager:", pager))

    output.append(("Using delimiter:", iocommands.get_current_delimiter()))
    output.append(("Using outfile:", iocommands.tee_file.name if iocommands.tee_file else ''))

    output.append(("Server version:", f'{global_variables["version"]} {global_variables["version_comment"]}'))
    output.append(("Protocol version:", global_variables["protocol_version"]))
    if cipher := get_ssl_cipher(cur):
        output.append(('SSL:', f'Cipher in use is {cipher}'))
    else:
        output.append(('SSL:', ''))
    output.append(('SSL/TLS version:', get_ssl_version(cur) or ''))

    if getattr(cur.connection, 'unix_socket', None):
        host_info = cur.connection.host_info
    else:
        host_info = f'{cur.connection.host} via TCP/IP'

    output.append(("Connection:", host_info))

    charset_spec = [
        {'name': 'Server characterset:', 'variable': 'character_set_server'},
        {'name': 'Db     characterset:', 'variable': 'character_set_database'},
        {'name': 'Client characterset:', 'variable': 'character_set_client'},
        {'name': 'Conn.  characterset:', 'variable': 'character_set_connection'},
        {'name': 'Result characterset:', 'variable': 'character_set_results'},
    ]
    for elt in charset_spec:
        if elt['variable'] in session_variables:
            value = session_variables[elt['variable']]
        else:
            value = ''
        output.append((elt['name'], value))

    if getattr(cur.connection, 'unix_socket', None):
        output.append(('UNIX socket:', global_variables['socket']))
    else:
        output.append(('TCP port:', cur.connection.port))

    output.append(('Server timezone:', get_server_timezone(global_variables)))
    output.append(('Local  timezone:', get_local_timezone()))

    if "Uptime" in status:
        output.append(("Uptime:", format_uptime(status["Uptime"])))

    if "Threads_connected" in status:
        # Print the current server statistics.
        stats = []
        stats.append(f'Connections: {status["Threads_connected"]}')
        if "Queries" in status:
            stats.append(f'Queries: {status["Queries"]}')
        stats.append(f'Slow queries: {status["Slow_queries"]}')
        stats.append(f'Opens: {status["Opened_tables"]}')
        if "Flush_commands" in status:
            stats.append(f'Flush tables: {status["Flush_commands"]}')
        stats.append(f'Open tables: {status["Open_tables"]}')
        if "Queries" in status:
            queries_per_second = int(status["Queries"]) / int(status["Uptime"])
            stats.append(f'Queries per second avg: {queries_per_second:.3f}')
        stats_str = "  ".join(stats)
        footer.append("\n" + stats_str)

    footer.append("--------------")

    return [SQLResult(preamble="\n".join(preamble), header=header, rows=output, postamble="\n".join(footer))]



from __future__ import annotations

from io import TextIOWrapper
import os
import sys
import time
from typing import TYPE_CHECKING

import click
import prompt_toolkit
from prompt_toolkit.shortcuts import ProgressBar
from prompt_toolkit.shortcuts.progress_bar import formatters as progress_bar_formatters
import pymysql

from mycli.packages.batch_utils import statements_from_filehandle
from mycli.packages.interactive_utils import confirm_destructive_query
from mycli.packages.sql_utils import is_destructive

if TYPE_CHECKING:
    from mycli.client import MyCli
    from mycli.main import CliArgs


class CheckpointReplayError(Exception):
    pass


def replay_checkpoint_file(
    batch_path: str,
    checkpoint: TextIOWrapper | None,
    resume: bool,
) -> int:
    if not resume:
        return 0

    if checkpoint is None:
        return 0

    if batch_path == '-':
        raise CheckpointReplayError('--resume is incompatible with reading from the standard input.')

    checkpoint_name = checkpoint.name
    checkpoint.flush()
    completed_count = 0
    try:
        with click.open_file(batch_path) as batch_h, click.open_file(checkpoint_name, mode='r', encoding='utf-8') as checkpoint_h:
            try:
                batch_gen = statements_from_filehandle(batch_h)
            except ValueError as e:
                raise CheckpointReplayError(f'Error reading --batch file: {batch_path}: {e}') from None
            for checkpoint_statement, _checkpoint_counter in statements_from_filehandle(checkpoint_h):
                try:
                    batch_statement, _batch_counter = next(batch_gen)
                except StopIteration:
                    raise CheckpointReplayError('Checkpoint script longer than batch script.') from None
                except ValueError as e:
                    raise CheckpointReplayError(f'Error reading --batch file: {batch_path}: {e}') from None
                if checkpoint_statement != batch_statement:
                    raise CheckpointReplayError(f'Statement mismatch: {checkpoint_statement}.')
                completed_count += 1
    except ValueError as e:
        raise CheckpointReplayError(f'Error reading --checkpoint file: {checkpoint.name}: {e}') from None
    except FileNotFoundError as e:
        raise CheckpointReplayError(f'FileNotFoundError: {e}') from None
    except OSError as e:
        raise CheckpointReplayError(f'OSError: {e}') from None

    return completed_count


def dispatch_batch_statements(
    mycli: 'MyCli',
    cli_args: 'CliArgs',
    statements: str,
    batch_counter: int,
) -> None:
    if batch_counter:
        if cli_args.format == 'csv':
            mycli.main_formatter.format_name = 'csv-noheader'
        elif cli_args.format == 'tsv':
            mycli.main_formatter.format_name = 'tsv_noheader'
        elif cli_args.format == 'table':
            mycli.main_formatter.format_name = 'ascii'
        else:
            mycli.main_formatter.format_name = 'tsv'
    else:
        if cli_args.format == 'csv':
            mycli.main_formatter.format_name = 'csv'
        elif cli_args.format == 'tsv':
            mycli.main_formatter.format_name = 'tsv'
        elif cli_args.format == 'table':
            mycli.main_formatter.format_name = 'ascii'
        else:
            mycli.main_formatter.format_name = 'tsv'

    warn_confirmed: bool | None = True
    if not cli_args.noninteractive and mycli.destructive_warning and is_destructive(mycli.destructive_keywords, statements):
        try:
            # this seems to work, even though we are reading from stdin above
            sys.stdin = open('/dev/tty')
            # bug: the prompt will not be visible if stdout is redirected
            warn_confirmed = confirm_destructive_query(mycli.destructive_keywords, statements)
        except (IOError, OSError) as e:
            mycli.logger.warning('Unable to open TTY as stdin.')
            raise e
    if warn_confirmed:
        if cli_args.throttle > 0 and batch_counter >= 1:
            time.sleep(cli_args.throttle)
        mycli.run_query(statements, checkpoint=cli_args.checkpoint, new_line=True)


def main_batch_with_progress_bar(mycli: 'MyCli', cli_args: 'CliArgs') -> int:
    goal_statements = 0
    if cli_args.batch is None:
        return 1
    if not sys.stdin.isatty() and cli_args.batch != '-':
        click.secho('Ignoring STDIN since --batch was also given.', err=True, fg='yellow')
    if os.path.exists(cli_args.batch) and not os.path.isfile(cli_args.batch):
        click.secho('--progress is only compatible with a plain file.', err=True, fg='red')
        return 1
    try:
        completed_statement_count = replay_checkpoint_file(cli_args.batch, cli_args.checkpoint, cli_args.resume)
        batch_count_h = click.open_file(cli_args.batch)
        for _statement, _counter in statements_from_filehandle(batch_count_h):
            goal_statements += 1
        batch_count_h.close()
        batch_h = click.open_file(cli_args.batch)
        batch_gen = statements_from_filehandle(batch_h)
    except (OSError, FileNotFoundError):
        click.secho(f'Failed to open --batch file: {cli_args.batch}', err=True, fg='red')
        return 1
    except ValueError as e:
        click.secho(f'Error reading --batch file: {cli_args.batch}: {e}', err=True, fg='red')
        return 1
    except CheckpointReplayError as e:
        name = cli_args.checkpoint.name if cli_args.checkpoint else 'None'
        click.secho(f'Error replaying --checkpoint file: {name}: {e}', err=True, fg='red')
        return 1
    try:
        if goal_statements:
            pb_style = prompt_toolkit.styles.Style.from_dict({'bar-a': 'reverse'})
            custom_formatters = [
                progress_bar_formatters.Bar(start='[', end=']', sym_a=' ', sym_b=' ', sym_c=' '),
                progress_bar_formatters.Text(' '),
                progress_bar_formatters.Progress(),
                progress_bar_formatters.Text(' '),
                progress_bar_formatters.Text('eta ', style='class:time-left'),
                progress_bar_formatters.TimeLeft(),
                progress_bar_formatters.Text(' ', style='class:time-left'),
            ]
            err_output = prompt_toolkit.output.create_output(stdout=sys.stderr, always_prefer_tty=True)
            with ProgressBar(style=pb_style, formatters=custom_formatters, output=err_output) as pb:
                for _pb_counter in pb(range(goal_statements)):
                    statement, statement_counter = next(batch_gen)
                    if statement_counter < completed_statement_count:
                        continue
                    dispatch_batch_statements(mycli, cli_args, statement, statement_counter)
    except (ValueError, StopIteration, IOError, OSError, pymysql.err.Error) as e:
        click.secho(str(e), err=True, fg='red')
        return 1
    finally:
        batch_h.close()
    return 0


def main_batch_without_progress_bar(mycli: 'MyCli', cli_args: 'CliArgs') -> int:
    if cli_args.batch is None:
        return 1
    if not sys.stdin.isatty() and cli_args.batch != '-':
        click.secho('Ignoring STDIN since --batch was also given.', err=True, fg='red')
    try:
        completed_statement_count = replay_checkpoint_file(cli_args.batch, cli_args.checkpoint, cli_args.resume)
        batch_h = click.open_file(cli_args.batch)
    except (OSError, FileNotFoundError):
        click.secho(f'Failed to open --batch file: {cli_args.batch}', err=True, fg='red')
        return 1
    except CheckpointReplayError as e:
        name = cli_args.checkpoint.name if cli_args.checkpoint else 'None'
        click.secho(f'Error replaying --checkpoint file: {name}: {e}', err=True, fg='red')
        return 1
    try:
        for statement, counter in statements_from_filehandle(batch_h):
            if counter < completed_statement_count:
                continue
            dispatch_batch_statements(mycli, cli_args, statement, counter)
    except (ValueError, StopIteration, IOError, OSError, pymysql.err.Error) as e:
        click.secho(str(e), err=True, fg='red')
        return 1
    finally:
        batch_h.close()
    return 0


def main_batch_from_stdin(mycli: 'MyCli', cli_args: 'CliArgs') -> int:
    batch_h = click.get_text_stream('stdin')
    try:
        for statement, counter in statements_from_filehandle(batch_h):
            dispatch_batch_statements(mycli, cli_args, statement, counter)
    except (ValueError, StopIteration, IOError, OSError, pymysql.err.Error) as e:
        click.secho(str(e), err=True, fg='red')
        return 1
    return 0
